from pathlib import Path
from contextlib import nullcontext

import aiohttp

from dan.core import aiofiles


async def fetch_file(url, dest: Path, name: str = None, chunk_size=1024, progress=None):
    if name is None:
        name = dest.name
    timeout = aiohttp.ClientTimeout(
        total=30 * 60, connect=30, sock_connect=30, sock_read=None
    )
    async with aiohttp.ClientSession(timeout=timeout) as session:
        async with session.get(url) as resp:
            if resp.status != 200:
                message = await resp.read()
                raise RuntimeError(f"unable to fetch {url}: {message.decode()}")
            size = int(resp.headers.get("content-length", 0))

            # with progress.Bar(
            #     f"downloading {name}", total=size // 1024, leave=False, unit="Ko"
            # ) as progressbar:
            with progress(f"downloading {name}", total=size // 1024) as bar:
                async with aiofiles.open(dest, mode="wb") as f:
                    async for chunk in resp.content.iter_chunked(chunk_size):
                        await f.write(chunk)
                        bar(len(chunk) // 1024)



class MakefileRegister:
    
    __makefile = None

    def __init_subclass__(cls, *args, internal=False, **kwargs):
        if not internal:
            from dan.core.include import context
            cls.__makefile = context.current
            cls.__makefile.register(cls)

    @classmethod
    def get_static_makefile(cls):
        return cls.__makefile

    @property
    def makefile(self) -> 'MakeFile':
        return self.__makefile
    
    @property
    def context(self):
        return self.__makefile.context
    
    @makefile.setter
    def makefile(self, value):
        assert self.__makefile is None, 'makefile should be set once'
        self.__makefile = value



from dan.core.target import Target, Installer
from dan.core.runners import async_run
from dan.core.find import find_executable, find_files
from dan.core.pm import re_match
from dan.core import asyncio, aiofiles
from dan.core.utils import chunks
from dan.cxx import Toolchain

from pathlib import Path
import os
import re

class Project(Target, internal=True):
    env: dict[str, str] = None
    configure_output: Path|str = None
    configure_options: list[str] = None
    make_options: list[str] = None

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.toolchain: Toolchain = self.context.get('cxx_target_toolchain')
        self.output = 'libav.built'
        self.__make = self.cache.get('make_path', None)
        
    @property
    def make(self):
        if self.__make is None:
            if self.toolchain.system.startswith('msys'):
                from dan.core.win import find_installation_data
                data = find_installation_data('MSYS2')
                self.__make = find_executable(r'.+make', data['InstallLocation'], default_paths=False)
                if self.__make is None:
                    raise RuntimeError('cannot find make')
                self.cache['make_path'] = self.__make
            else:
                self.__make = 'make'
        return self.__make

    def __get_make_args(self):
        make_args = []
        if self.make_options is not None:
            make_args.extend(self.make_options)
        return make_args
    
    def __get_env(self):
        env = dict(self.toolchain.env)
        if self.env is not None:
            for k, v in self.env.items():
                env[k] = v
        env['PKG_CONFIG_PATH'] = str(self.makefile.root.pkgs_path)
        env['CC'] = str(self.toolchain.cc)
        env['CXX'] = str(self.toolchain.cxx)
        return env

    async def __build__(self):
        if not self.toolchain.system.is_linux and not self.toolchain.system.startswith('msys'):
            raise RuntimeError(f'{self.name} can only be built on linux or msys2-mingw')

    async def do_compile(self, obj, options, env):
        self.info('generating %s', obj)
        await async_run(f'{self.toolchain.cc} {options}', logger=self, cwd=self.build_path, env=env)
        self.info('%s generated', obj)
        
    async def do_ar(self, lib, ar_arg, objects: list[str], builds, env):
        self.info('generating %s objects', lib)

        async with asyncio.TaskGroup(f'generating {lib} objects') as obj_builds:
            for obj in objects:
                if obj in builds:
                    obj_builds.create_task(builds[obj])

        command = f'{self.toolchain.ar} {ar_arg} {lib} {" ".join(objects)}'
        windows_command_line_limitation = 8191
        if os.name == 'nt' and len(command) >= windows_command_line_limitation:
            # workaround command-line size limitation
            objects_dir = self.build_path / f'{lib}_objects'
            if objects_dir.exists():
                await aiofiles.rmtree(objects_dir)
            objects_dir.mkdir()
            for obj in objects:
                obj_path = Path(self.build_path / obj)
                dest = objects_dir / obj_path.relative_to(objects_dir.parent).as_posix().replace('/', '_')
                if not dest.exists():
                    obj_path.rename(dest)
                
            command = f'{self.toolchain.ar} {ar_arg} {lib} {(objects_dir.relative_to(self.build_path) / "*.o").as_posix()}'
                
        self.info('creating static library %s', lib)
        await async_run(command, logger=self, cwd=self.build_path, env=env)
        self.info('static library %s created', lib)

    async def do_ld(self, output, lib_paths, objects, libs, builds: dict[str, asyncio.Task], env):
        self.info('generating %s dependencies', output)

        async with asyncio.TaskGroup(f'generating {output} objects') as grp:
            for obj in objects:
                if obj in builds:
                    grp.create_task(builds[obj])

        async with asyncio.TaskGroup(f'generating {output} dependencies') as grp:
            seen = set()
            for lib in re.split(r'\s+', libs):
                lib = lib[2:]
                if lib not in seen:
                    seen.add(lib)
                    for l in builds:
                        if re.search(rf'lib{lib}\.(a|so|dll|lib)', l):
                            grp.create_task(builds[l])

        self.info('linking %s', output)
        await async_run(f'{self.toolchain.cc} {lib_paths} -o {output} {" ".join(objects)} {libs}', logger=self, cwd=self.build_path, env=env)
        self.info('%s linked', output)
    
    @staticmethod
    async def patch_makefile(makefile: Path):
        from dan.core.win import cygpath
        bak = makefile.with_suffix('.bak')
        bak.unlink(missing_ok=True)
        makefile.rename(bak)
        async with aiofiles.open(bak) as i, aiofiles.open(makefile, 'w') as o:
            for line in await i.readlines():
                m = re.match(r'(.+?[= ])(/.+)', line)
                if m:
                    line = m[1] + cygpath(m[2], reverse=True) + '\n'
                    await o.write(line)
                else:
                    await o.write(line)


    async def __install__(self, installer: Installer):
        config_options = []
        if self.configure_options is not None:
            config_options.extend(self.configure_options)
        config_options.append(f'--prefix={installer.settings.destination}')

        env = self.__get_env()
        await async_run(['bash', self.source_path / 'configure', *config_options], cwd=self.build_path, logger=self, env=env)

        if not self.toolchain.system.is_linux:
            # patch MakeFiles
            makefile = Path(self.build_path / 'Makefile')
            if makefile.exists():
                await self.patch_makefile(makefile)
            
            for makefile in find_files(r'.+\.mak', self.build_path):
                await self.patch_makefile(makefile)

        builds = dict()
        installs = list()

        async with asyncio.TaskGroup(f'executing make {self.name}') as g:
            async def wrap_executor(stream):
                with stream as lines:
                    async for line in lines:
                        line = line.strip()
                        match re_match(line):
                            case r'.+CC.+? (.+\.o); (.+?) (.+)' as m:
                                obj = m[1]
                                # cc = m[2]
                                options = m[3]
                                builds[obj] = asyncio.create_task(self.do_compile(obj, options, env))

                            case r'.+AR.+? (.+\.a); (.+?) (.+?) (.+\.a) (.+)' as m:
                                lib = m[1]
                                # ar = m[2]
                                ar_arg = m[3]
                                objects = m[5].split(' ')
                                builds[lib] = asyncio.create_task(self.do_ar(lib, ar_arg, objects, builds, env))

                            case r'.+LD.+? (.+); (.+?) (.+?)\s?-o (.+?) (.+?) (-l.+)' as m:
                                lib_paths = m[3]
                                output = m[4]
                                objects = m[5].split(' ')
                                libs = m[6]
                                builds[output] = asyncio.create_task(self.do_ld(output, lib_paths, objects, libs, builds, env))
                                
                            case r'.+INSTALL.+? (.+?); install (.+(?=-m \d+))?(-m \d+) (.+)' as m:
                                # args = m[2]
                                # mode = m[3]
                                items = str(m[4]).split(' ')
                                dest = items.pop()
                                if dest[0] in ('"', "'"):
                                    dest = dest[1:-1] # unquote
                                dest = Path(dest).expanduser()
                                for item in items:
                                    installs.append(installer._install(self.build_path / item, dest))
                            case _:
                                # by default we assume it will be a bash command
                                await async_run(['bash', '-c', line], logger=self, cwd=self.build_path)

            await async_run([self.make, 'install', '-n', *self.__get_make_args()],
                             cwd=self.build_path, logger=self, log=False,
                             env=env, out_capture=wrap_executor)
        
        async with asyncio.TaskGroup(f'generating {self.name} targets') as g:
            for name, coro in builds.items():
                g.create_task(coro)
            
        for chunk in chunks(installs, 100):
            await asyncio.gather(*chunk)

        await super().__install__(installer)
        self.output.touch()



import os
import pathlib

Path = type(pathlib.Path())

@property
def modification_time(self):
    return self.stat().st_mtime

Path.modification_time = modification_time


def younger_than(self, other):
    time = other if isinstance(other, float) else other.modification_time
    assert isinstance(time, float)
    return self.modification_time > time

Path.younger_than = younger_than


def older_than(self, other):
    time = other if isinstance(other, float) else other.modification_time
    assert isinstance(time, float)
    return self.modification_time < time

Path.older_than = older_than


def utime(self, *args, **kw_args):
    os.utime(self, *args, **kw_args)

Path.utime = utime

@property
def is_empty(self):
    if os.path.exists(self) and not os.path.isfile(self):

        # Checking if the directory is empty or not
        if not os.listdir(self):
            return True
        else:
            return False
    else:
        return False

Path.is_empty = is_empty



import atexit
import math
import time
import typing as t
import sys
import shutil
import weakref
import logging
import enum
from contextlib import nullcontext
from dan.core import asyncio

from termcolor import colored

from dan.core import asyncio


class TerminalMode(enum.Enum):
    BASIC = 1
    STICKY = 2
    CODE = 3


mode = TerminalMode.STICKY


def set_mode(new_mode: TerminalMode):
    global mode
    global _manager
    if _manager is not None:
        raise RuntimeError("Cannot change terminal mode once initialized")
    mode = new_mode


class TermSequence:
    """Terminal sequence codes

    :see: https://gist.github.com/fnky/458719343aabd01cfb17a3a4f7296797
    """

    ESC = "\x1b"
    SEQ_START = ESC + "["

    @staticmethod
    def current_pos():
        return "\x1b[6n"

    @staticmethod
    def home():
        """Moves cursor to home position (0, 0)"""
        return f"\x1b[H"

    @staticmethod
    def move(line, colomn=0):
        """Moves cursor down"""
        return f"\x1b[{line};{colomn}H"

    @staticmethod
    def up(n=1):
        """Moves cursor up"""
        return f"\x1b[{n}A"

    @staticmethod
    def down(n=1):
        """Moves cursor down"""
        return f"\x1b[{n}B"

    @staticmethod
    def scroll_up(n=1):
        """Scroll viewport up"""
        return f"\x1b[{n}S"

    @staticmethod
    def scroll_down(n=1):
        """Scroll viewport up"""
        return f"\x1b[{n}T"

    @staticmethod
    def next(n=1):
        """Moves cursor to next line"""
        return f"\x1b[{n}E"

    @staticmethod
    def prev(n=1):
        """Moves cursor to previous line"""
        return f"\x1b[{n}F"

    @staticmethod
    def left(n=1):
        """Moves cursor left"""
        return f"\x1b[{n}D"

    @staticmethod
    def right(n=1):
        """Moves cursor left"""
        return f"\x1b[{n}C"

    @staticmethod
    def hide_cursor():
        """Makes cursor invisible"""
        return f"\x1b[?25l"

    @staticmethod
    def show_cursor():
        """Makes cursor visible"""
        return f"\x1b[?25h"

    class hidden_cursor:
        def __enter__(self):
            sys.stdout.write(TermSequence.hide_cursor())

        def __exit__(self, *args):
            sys.stdout.write(TermSequence.show_cursor())

    @staticmethod
    def line_clear(n=2):
        """Erase line

        n == 0: erase from cursor to end of line
        n == 1: erase start of line to the cursor
        n == 2: erase the entire line
        """
        return f"\x1b[{n}K"

    __next_clear = next() + line_clear()

    @staticmethod
    def next_clear():
        """Moves to next line and clear it"""
        return TermSequence.__next_clear

    @staticmethod
    def sreen_clear(n=2):
        """Erase screen

        n == 0: erase from cursor until end of screen
        n == 1: erase from cursor to beginning of screen
        n == 2: erase entire screen
        n == 3: erase saved lines
        """
        return f"\x1b[{n}J"


class _OutputStreamProgress:
    UTF = " " + "".join(map(chr, range(0x258F, 0x2587, -1)))

    def __init__(
        self, stream: "TermStream", status: str, total=None, auto_update=True
    ) -> None:
        self._s = stream
        self._status = status
        self.total = total
        self.n = 0
        self._saved_status = None
        self._saved_extra = None
        self._elapsed_s = 0.0
        self._auto_update = auto_update
        self._auto_update_task = None
        self._done = False

    async def __auto_update(self):
        while True:
            await asyncio.sleep(0.25)
            self.__call__(0)

    def __enter__(self):
        self._done = False
        self._t_start = time.time()
        self._saved_status = self._s._status
        self._saved_extra = self._s._extra
        self._s._status = self._status
        self._s._extra = self
        self._s.visible = True
        if self._auto_update:
            self._auto_update_task = asyncio.create_task(self.__auto_update())
        return self

    def __exit__(self, *args):
        self._done = True
        if self._auto_update_task is not None:
            self._auto_update_task.cancel()
            self._auto_update_task = None
        
        if mode == TerminalMode.CODE:
            # we want done/done to be printed
            def reset():
                self._s._status = self._saved_status
                self._s._extra = self._saved_extra
                self._s.update()
            self._s._mngr._flush_callbacks.append(reset)
        else:
            self._s._status = self._saved_status
            self._s._extra = self._saved_extra
        self._s.update()

    def __call__(self, n=1, status=None):
        self._elapsed_s = time.time() - self._t_start
        self.n += n
        if status is not None:
            self._s._status = self._status = status
        self._s.update()

    def __make_bar(self, frac, width):
        nsyms = len(self.UTF) - 1
        if frac is None:
            cursor_width = 3
            width -= cursor_width
            pos = 0.5 + math.sin(math.pi * self._elapsed_s) / 2
            offset, _ = divmod(pos * width * nsyms, nsyms)
            bar = (
                " " * int(offset)
                + self.UTF[-1] * cursor_width
                + " " * int(width - offset)
            )
        else:
            bar_length, frac_bar_length = divmod(int(frac * width * nsyms), nsyms)
            bar = self.UTF[-1] * bar_length
            if bar_length < width:  # whitespace padding
                bar = (
                    bar
                    + self.UTF[frac_bar_length]
                    + self.UTF[0] * (width - bar_length - 1)
                )
        return bar

    def __str_code(self) -> str:
        if self._done:
            return 'done/done'
        return f'{self.n}/{self.total}'

    def __str_default(self) -> str:
        if self.total:
            frac = self.n / self.total
            rbar = f"{int(100 * frac):3d}%"
            eta = (1 - frac) * self._elapsed_s
            rbar += f" - eta: {int(eta)}s"
        else:
            frac = None
            rbar = "   ∞"
        bar = self.__make_bar(frac, self._s._mngr.width // 5)
        return f" |{bar}| {rbar}"

    def __str__(self) -> str:
        match mode:
            case TerminalMode.CODE:
                return self.__str_code()
            case _:
                return self.__str_default()


class _TaskGroup(asyncio.TaskGroup, _OutputStreamProgress):
    def __init__(self, stream: "TermStream", name: str, **kwargs):
        _OutputStreamProgress.__init__(self, stream, status=name, **kwargs)
        asyncio.TaskGroup.__init__(self, name)

    def __notify_bar_task_done(self, task):
        self._elapsed_s = time.time() - self._t_start
        self.n += 1
        self._s.update()

    async def __aexit__(self, et, exc, tb):
        total = len(self._tasks)
        if total:
            self.total = total
            for task in self._tasks:
                task.add_done_callback(self.__notify_bar_task_done)
            self._s.update()
        _OutputStreamProgress.__enter__(self)
        result = await asyncio.TaskGroup.__aexit__(self, et, exc, tb)
        _OutputStreamProgress.__exit__(self, et, exc, tb)
        # async def delayed_exit():
        #     await asyncio.sleep(0.5)
        #     _OutputStreamProgress.__exit__(self, et, exc, tb)
        # asyncio.spawn(delayed_exit())
        return result


class ColorTheme:
    def __init__(self, color: str, attrs: t.Iterable[str] = None) -> None:
        self.color = color
        self.attrs = attrs

    def __call__(self, s: str) -> t.Any:
        return colored(s, self.color, attrs=self.attrs)


class TermLogHandler(logging.Handler):
    def emit(self, record):
        write(self.format(record))


class TermStreamColorTheme:
    default_icon = ColorTheme("grey", ["bold"])
    default_name = ColorTheme("cyan", ["bold"])
    default_status = ColorTheme("grey", ["bold"])
    default_extra = ColorTheme("white")

    def __init__(
        self,
        icon: ColorTheme = None,
        name: ColorTheme = None,
        status: ColorTheme = None,
        extra: ColorTheme = None,
    ) -> None:
        self.icon = icon or self.default_icon
        self.name = name or self.default_name
        self.status = status or self.default_status
        self.extra = extra or self.default_extra


default_theme = TermStreamColorTheme()


class TermStream:
    DEFAULT_ICON = "➔"
    CHILD_ICON = "➥"

    def __init__(
        self, name: str, theme: TermStreamColorTheme = None, parent: "TermStream" = None
    ) -> None:
        self.name = name
        self._visible = False
        self._dirty = False
        self._icon = self.DEFAULT_ICON
        self._status: str = ""
        self._extra: str = ""
        self.theme = theme or default_theme
        self._cached_out: list[str] = list()
        self._mngr = manager()
        self._hide_task: asyncio.Task = None
        self._offset = 0
        self._parent = parent
        self._children: list[weakref.ReferenceType[TermStream]] = list()
        self._weak_self = weakref.ref(self)
        if self._parent is not None:
            self._offset = self._parent._offset + 2
            self._parent._children.append(self._weak_self)
            self._icon = self.CHILD_ICON
        else:
            self._mngr._streams.append(self._weak_self)

        match mode:
            case TerminalMode.CODE:
                self._get_output = self._get_output_code
                from dan.logging import _no_color_formatter
                self._formatter = _no_color_formatter
            case _:
                self._get_output = self._get_output_default

    def __del__(self):
        if self._parent:
            self._parent._children.remove(self._weak_self)
            self._parent.update()
        else:
            self._mngr._streams.remove(self._weak_self)
            self._mngr.update()

    def sub(self, name, theme: TermStreamColorTheme = None):
        return TermStream(name, theme, parent=self)

    @property
    def prefix_width(self):
        return self._mngr.width - (
            len(self._icon) + 1 + len(self.name) + 2 + self._offset
        )

    @property
    def visible(self):
        return self._visible

    @visible.setter
    def visible(self, value):
        self._visible = value
        self._dirty = True

    def hide(self):
        self.visible = False
        self.update()

    def hide_children(self):
        for ref in self._children:
            ref().hide()
        self.update()

    async def __hide_after(self, delay):
        await asyncio.sleep(delay)
        self.hide()

    def __hide_task_done(self, t: asyncio.Task):
        if t.cancelled():
            self.hide()

    def status(self, msg: str, icon: str = None, timeout=None):
        self._status = msg
        if icon is not None:
            self._icon = icon
        self._visible = True
        if self._hide_task is not None:
            self._hide_task.remove_done_callback(self.__hide_task_done)
            self._hide_task.cancel()
            self._hide_task = None
        if timeout is not None:
            self._hide_task = asyncio.create_task(self.__hide_after(timeout))
            self._hide_task.add_done_callback(self.__hide_task_done)
        self.update()

    def update(self):
        self._dirty = True
        if not self._status:
            self._visible = False
        if self._parent is not None:
            self._parent.update()
        else:
            self._mngr.update()

    def _refresh_state(self):
        return self._dirty, self._visible

    def progress(self, desc, total=None, **kwargs):
        return _OutputStreamProgress(self, desc, total=total, **kwargs)

    def task_group(self, name):
        return _TaskGroup(self, name)

    ts = TermSequence

    def _get_output_code(self, now) -> list[str]:
        if self._dirty:
            from dan.logging import STATUS, PROGRESS
            if isinstance(self._extra, _OutputStreamProgress):
                record = logging.LogRecord(self.name, PROGRESS, '', 0, f'{self._status} - {self._extra}', None, None)
            else:
                record = logging.LogRecord(self.name, STATUS, '', 0, f'{self._status}', None, None)

            self._cached_out = [self._formatter.format(record) + '\n']
            for ref in self._children:
                child = ref()
                if child.visible:
                    self._cached_out.extend(child._get_output(now))
            self._dirty = False
        return self._cached_out

    def _get_output_default(self, now) -> list[str]:
        if self._dirty:
            prefix = f"{' ' * self._offset}{self.theme.icon(self._icon)}  {self.theme.name(self.name)}: "
            extra = str(self._extra)
            status = self._status.replace('\n', ' ') # TODO handle multiline status
            max_status_len = self.prefix_width - (len(extra)) - 1
            if len(status) > max_status_len:
                status = status[: max_status_len - 4] + " ..."
            if len(extra):
                padding = self.prefix_width - len(status) - len(extra) - 1
                padding = padding * " "
                suffix = f"{padding}{self.theme.extra(extra)}"
            else:
                suffix = ""
            self._cached_out = [f"{prefix}{status}{suffix}\n"]
            for ref in self._children:
                child = ref()
                if child.visible:
                    self._cached_out.extend(child._get_output(now))
            self._dirty = False
        return self._cached_out


class _TermManager:
    def __init__(self, min_update_freq=1, max_update_freq=12) -> None:
        self._streams: list[weakref.ReferenceType[TermStream]] = list()
        self._fp = sys.stdout
        # self._fp.reconfigure(write_through=False)
        self._min_delay = 1 / min_update_freq
        self._max_delay = 1 / max_update_freq
        self._up_ev = asyncio.Event()
        self._stop_requested = False
        self._raw_lines = list()
        self._thread = None
        self._flush_callbacks = list()

        global _manager
        if _manager is not None:
            raise RuntimeError("Only one _TermManager can be created")

        match mode:
            case TerminalMode.CODE:
                self._flush = self._flush_code
            case _:
                self._flush = self._flush_sticky

    def __del__(self):
        self.stop()

    @property
    def height(self):
        return shutil.get_terminal_size()[1]

    @property
    def width(self):
        return shutil.get_terminal_size()[0]

    def start(self):
        if self._thread is not None:
            self.stop()
        try:
            loop = asyncio.get_running_loop()
            self._thread = asyncio.create_task(self._render(), name="terminal-rendering")
        except RuntimeError:
            pass

    def stop(self):
        if self._thread is not None:
            self._stop_requested = True
            self._up_ev.set()

    def update(self):
        self._up_ev.set()

    def write(self, s: str, end: str = "\n"):
        self._raw_lines.append(s + end)
        self.update()

    ts = TermSequence

    def _flush_sticky(self, prev_line_count):
        out = self._fp
        self._up_ev.clear()
        stop = self._stop_requested
        now = time.time()
        output_lines = []
        if prev_line_count:
            output_lines.append(self.ts.prev(prev_line_count))
        prev_line_count = 0
        force = len(self._raw_lines)
        max_height = self.height - 2

        status_lines = []
        if mode == TerminalMode.STICKY:
            for ref in self._streams:
                stream = ref()
                if stream is None:
                    continue
                is_dirty, is_visible = stream._refresh_state()
                if is_dirty or force:
                    prev_height = len(stream._cached_out)
                    status_lines.append(self.ts.line_clear())
                    data = stream._get_output(now)
                    height = len(data)
                    if prev_line_count + height > max_height:
                        break

                    if is_visible:
                        if prev_height != height:
                            force = True  # force re-render next streams
                        status_lines.extend(data)
                        prev_line_count += height
                    elif prev_height != 1:
                        force = True
                elif is_visible:
                    # just jump over
                    height = len(stream._cached_out)
                    if prev_line_count + height > max_height:
                        break

                    prev_line_count += height
                    status_lines.append(self.ts.next(height))

            # insert separator
            if prev_line_count:
                status_lines.insert(0, "―" * (self.width) + "\n")
                prev_line_count += 1

        if self._raw_lines:
            lines = []
            line_count = 0
            for ls in self._raw_lines:
                nls = [l.rstrip() + "\n" for l in ls.splitlines()]
                line_count += len(nls)
                for l in nls:
                    line_count += len(l) // self.width
                lines.extend(nls)
            if mode == TerminalMode.STICKY:
                output_lines.append(self.ts.sreen_clear(0))
                if line_count + prev_line_count > max_height:
                    output_lines.append(
                        self.ts.scroll_up(prev_line_count)
                        + self.ts.prev(prev_line_count)
                    )
            output_lines.extend(lines)

            self._raw_lines = list()

        if mode == TerminalMode.STICKY:
            # clear rest of the screen
            status_lines.append(self.ts.sreen_clear(0))

        # for line in [*output_lines, *status_lines]:
        #     out.write(line)
        #     out.flush()
        out.writelines([*output_lines, *status_lines])
        out.flush()
        
        for callback in self._flush_callbacks:
            callback()
        self._flush_callbacks = list()

        return stop, prev_line_count
    
    def _flush_code(self, prev_line_count):
        out = self._fp
        self._up_ev.clear()
        stop = self._stop_requested
        now = time.time()
        if self._raw_lines:
            out.writelines(self._raw_lines)
            self._raw_lines = []
        for ref in self._streams:
            stream = ref()
            if stream is None:
                continue
            is_dirty, is_visible = stream._refresh_state()
            if is_dirty and is_visible:
                data = stream._get_output(now)
                out.writelines(data)
        
        for callback in self._flush_callbacks:
            callback()
        self._flush_callbacks = list()
            
        return stop, 0

    async def _render(self):
        out = self._fp
        if mode == TerminalMode.STICKY:
            out.write(self.ts.home() + self.ts.sreen_clear())
        prev_line_count = 0
        stop = False
        last_update = 0

        async def auto_refresh():
            await asyncio.sleep(self._min_delay)
            self.update()

        auto_refresh_task = asyncio.create_task(auto_refresh())
        with self.ts.hidden_cursor() if mode == TerminalMode.STICKY else nullcontext():
        # with nullcontext():
            while not stop:
                await self._up_ev.wait()
                now = time.time()
                delay = now - last_update
                if delay < self._max_delay and not self._stop_requested:
                    await asyncio.sleep(self._max_delay - delay)
                stop, prev_line_count = self._flush(prev_line_count)
                last_update = time.time()

            auto_refresh_task.cancel()

            self._up_ev.set()
            while self._up_ev.is_set():
                _, prev_line_count = self._flush(prev_line_count)
                await asyncio.sleep(0)

        self._thread = None


_manager: _TermManager = None


def manager():
    global _manager
    if _manager is None:
        _manager = _TermManager()
        _manager.start()
        atexit.register(_cleanup_manager)
    return _manager


def _cleanup_manager():
    global _manager
    if _manager is not None:
        _manager.stop()
        del _manager


def write(s: str, end="\n"):
    return manager().write(s, end)



from dan import logging

import fnmatch
import os
import contextlib
from pathlib import Path

from dan.cli import click
from dan.core.requirements import parse_package
from dan.core.cache import Cache
from dan.core.runners import async_run
from dan.io.repositories import RepositoriesSettings, RepositoryConfig, _get_settings
from dan.make import Make
from dan.cxx.detect import get_dan_path
from dan.core import asyncio, aiofiles

def get_source_path():
    source_path = get_dan_path() / 'deps'
    source_path.mkdir(exist_ok=True, parents=True)
    return source_path

_make : Make = None
async def get_make(toolchain='default', quiet=True):
    global _make
    if _make is None:
        source_path = get_source_path()
        os.chdir(source_path)
        (source_path / 'dan-build.py').touch()
        kwds = dict()
        if quiet:
            kwds['verbose'] = -1
        make = Make(source_path / 'build', **kwds)
        make.config.source_path = str(source_path)
        make.config.build_path = str(source_path / 'build')
        make.config.toolchain = toolchain
        await make._config.save()
        await make.initialize()
        _make = make
    return _make


_repositories = None
async def get_repositories():
    global _repositories
    if _repositories is None:
        from dan.io.repositories import get_all_repo_instances
        await get_make()
        _repositories = get_all_repo_instances()
        async with asyncio.TaskGroup() as g:
            for repo in _repositories:
                g.create_task(repo.build())
    return _repositories

async def get_repository(name = None):
    from dan.io.repositories import get_repo_instance
    await get_make()
    repo = get_repo_instance(name)
    await repo.build()
    return repo


@contextlib.asynccontextmanager
async def make_context(toolchain='default', quiet=True):
    make = await get_make(toolchain, quiet=quiet)
    with make.context:
        yield make


@click.group()
@click.option('--verbose', '-v', count=True)
def cli(verbose):
    if verbose == 0:
        logging.getLogger().setLevel(logging.ERROR)
    elif verbose == 1:
        logging.getLogger().setLevel(logging.DEBUG)
    else:
        logging.getLogger().setLevel(logging.TRACE)

@cli.command()
@click.option('--setting', '-s', 'settings',
              help='Apply repository setting',
              type=click.SettingsParamType(RepositoriesSettings), multiple=True)
async def configure(settings):
    """Configure repository settings"""
    io_settings = _get_settings()
    from dan.core.settings import apply_settings
    apply_settings(io_settings, *settings, logger=click.logger)
    await Cache.save_all()

@cli.group()
def ls():
    """Inspect stuff"""
    pass


@ls.command()
async def repositories():
    """List available repositories"""
    async with make_context():
        repos = await get_repositories()
        for repo in repos:
            click.echo(repo.name)

@ls.command()
async def libraries():
    """List available libraries"""
    async with make_context():
        repos = await get_repositories()
        for repo in repos:
            for name, lib in repo.installed.items():
                click.echo(f'{name} = {lib.version}')

async def get_library(library_spec):
    package, library, repository = parse_package(library_spec)
    repo = await get_repository(repository)
    if repo is None:
        raise RuntimeError(f'cannot find repository {repository}')

    lib = repo.find(library, package)
    if lib is None:
        if repository is None:
            repository = repo.name
        if package is None:
            package = library
        raise RuntimeError(f'cannot find {package}:{library}@{repository}')

    return lib

@ls.command()
@click.argument('LIBRARY')
async def versions(library: str):
    """Get LIBRARY's available versions"""
    async with make_context():
        lib = await get_library(library)
        
        from dan.src.github import GitHubReleaseSources
        
        sources: GitHubReleaseSources = lib.get_dependency(GitHubReleaseSources)
        available_versions = await sources.available_versions()
        available_versions = sorted(available_versions.keys())
        for v in available_versions:
            if v == lib.version:
                click.echo(f' - {v} (default)')
            else:
                click.echo(f' - {v}')

@ls.command()
@click.argument('LIBRARY')
async def options(library: str):
    """Get LIBRARY's available options"""
    async with make_context():
        lib = await get_library(library)
        await lib.initialize()
        for o in lib.options:
            current = ''
            if o.value != o.default:
                current = f', current: {o.value}'
            click.echo(f'{o.name}: {o.help} (type: {o.type.__name__}, default: {o.default}{current})')

@cli.command()
@click.argument('NAME')
async def search(name):
    """Search for NAME in repositories"""
    async with make_context():
        name = f'*{name}*'
        repos = await get_repositories()
        for repo in repos:
            installed = repo.installed
            for libname, lib in installed.items():
                if fnmatch.fnmatch(libname, name):
                    click.echo(f'{libname} = {lib.version}')

@cli.command()
@click.option('--toolchain', '-t', type=click.ToolchainParamType(), default='default')
@click.argument('PACKAGE_SPEC')
@click.argument('VERSION', required=False)
async def install(toolchain, package_spec, version):
    """Intall given PACKAGE_SPEC"""
    from dan.io.package import PackageBuild

    async with make_context(toolchain, quiet=False) as make:
        package, name, repository = parse_package(package_spec)
        pkg = PackageBuild(name, version, package, repository, makefile=make.root)
        await pkg.initialize()
        if pkg.up_to_date:
            click.echo(f'Package {package_spec} already installed at version {pkg.version}')
        else:
            await pkg.build()
            click.echo(f'Package {package_spec} installed successfully at version {pkg.version}')

@cli.group()
def dev():
    """Developper utils"""
    pass

@dev.command()
@click.option('--force', '-f', is_flag=True)
@click.argument('NAME')
async def create_repository(force, name):
    """Create a new package repository"""
    package_path: Path = get_dan_path() / 'repositories' / name
    if package_path.exists():
        if force or click.confirm(f'{name} repository already exists, remove it ?', default=False):
            click.logger.info('Removing existing directory %s', package_path)
            await aiofiles.rmtree(package_path)
        else:
            return -1
    click.logger.info('Create directory %s', package_path)
    package_path.mkdir()
    out, err, rc = await async_run('git init .', cwd=package_path, log=False, logger=click.logger)
    click.logger.info(out.strip())

    import jinja2
    templateEnv = jinja2.Environment(loader=jinja2.PackageLoader('dan.cli'))
    for template_name in templateEnv.list_templates(filter_func=lambda x: x.startswith('package_repository/') and not x.startswith('__')):
        dest = package_path.absolute() / template_name.removeprefix('package_repository/')
        if not dest.parent.exists():
            click.logger.debug('Creating directory %s...', dest)
            dest.parent.mkdir(parents=True)
        async with aiofiles.open(dest, 'w') as f:
            click.logger.info('Generating %s...', dest)
            await f.write(templateEnv.get_template(template_name).render(
                name = name,
            ))
        
    io_settings = _get_settings()
    if name in io_settings.repositories:
        click.logger.info('Unregistering existing %s', name)

    click.logger.info('Registering %s (with no remote)', name)
    io_settings.repositories.append(RepositoryConfig(name, url=package_path.as_uri()))
    
    await Cache.save_all()

@dev.command()
@click.option('--force', '-f', is_flag=True)
@click.option('--repo-name', prompt='Local repository name')
@click.option('--package-name', prompt='Package name')
@click.option('--package-description', prompt='Package description')
@click.option('--package-username', prompt='Github repo owner username')
@click.option('--package-projectname', prompt='Github repo project name')
@click.option('--default-version', prompt='Default github release version')
@click.option('package_requirements', '--package-requirement', multiple=True)
async def create_package(force, repo_name, package_name, **kwargs):
    repo_path = get_dan_path() / 'repositories' / repo_name
    if not repo_path.exists():
        click.logger.error('%s does not exist', repo_path)
        return -1
    packages_path: Path = get_dan_path() / 'repositories' / repo_name / 'packages'
    dest = packages_path / package_name
    if dest.exists():
        if force or click.confirm(f'{dest} package folder already exists, remove it ?', default=False):
            click.logger.info('Removing existing directory %s', dest)
            await aiofiles.rmtree(dest)
        else:
            return -1
    dest.mkdir()

    import jinja2
    templateEnv = jinja2.Environment(loader=jinja2.PackageLoader('dan.cli'))
    async with aiofiles.open(dest / 'dan-build.py', 'w') as f:
        click.logger.info('Generating %s...', dest)
        await f.write(templateEnv.get_template('package/dan-build.py').render(
            package_name = package_name,
            **kwargs
        ))
    
    packages = [d.name for d in packages_path.iterdir() if d.is_dir() and (d / 'dan-build.py').exists()]
    async with aiofiles.open(packages_path / 'dan-build.py', 'w') as f:
        await f.write(templateEnv.get_template('package_repository/packages/dan-build.py').render(
            name = repo_name,
            packages = packages,
        ))


def main():
    import sys
    try:
        sys.exit(cli(auto_envvar_prefix='DAN'))
    except Exception as err:
        click.logger.error(str(err))
        _ex_type, _ex, tb = sys.exc_info()
        import traceback
        click.logger.debug(' '.join(traceback.format_tb(tb)))
        try:
            # wait asyncio loop to terminate
            asyncio.get_running_loop().run_until_complete()
        except Exception:
            pass
        return -1

if __name__ == '__main__':
    main()



import asyncio
import os
import re
import typing as t

import aiofiles
from dan.core import asyncio
from dan.core.pathlib import Path
from dan.core.register import MakefileRegister
from dan.logging import Logging


class AsyncExecutable(Logging):
    async def execute(self, *args, **kwargs): ...


class Case:
    def __init__(self, name, *args, expected_result: int = 0, expected_output: str = None, file=None, lineno=None, strip_output=True, normalize_newlines=True):
        self.name = name
        self.args = args
        self.expected_result = expected_result
        self.file = file
        self.lineno = lineno
        self.expected_output = expected_output
        self.strip_output = strip_output
        self.normalize_newlines = normalize_newlines


class Test(Logging, MakefileRegister, internal=True):
    """Test definition
    """

    name: str = None
    fullname: str = None

    cases: t.Iterable[Case] = [Case(None, expected_result=0)]
    """Test cases
    
    list of args giving a return value
    """

    executable: type[AsyncExecutable] = None

    file: Path | str = None
    lineno: int = None

    def __init__(self, *args, **kwargs):
        # in case of inheritance usage, we must initialize the AsyncExecutable part
        super().__init__(*args, **kwargs)

        if self.name is None:
            self.name = self.__class__.__name__

        if not self.executable:
            if not hasattr(self, 'execute'):
                raise RuntimeError(
                    f'Test "{self.name}" does not have executable set, if you intent to use it through inheritance, make sure "Test" is the first derived class')
            self.executable = self
        elif isinstance(self.executable, type):
            self.executable = self.makefile.find(self.executable)

        self.name = self.name or self.executable.name
        self.fullname = f'{self.executable.makefile.fullname}.{self.name}'
        self.file = Path(self.file) if self.file else None
        self.workingDir = self.executable.build_path

    def basename(self, caze: Case):
        if len(self) <= 1:
            return self.name
        else:
            return f'{self.name}.{caze.name}'

    def outs(self, caze: Case):
        base = self.basename(caze)
        base = re.sub(r'[^\w_. -]', '_', base)
        out = self.workingDir / f'{base}.stdout'
        err = self.workingDir / f'{base}.stderr'
        return out, err

    async def _run_test(self, caze: Case):
        name = f'{self.name}.{caze.name}' if caze is not None else self.name
        args = [str(a) for a in caze.args]
        self.debug('testing %s', name)
        out, err, rc = await self.executable.execute(*args, no_raise=True, cwd=self.workingDir)
        out_log, out_err = self.outs(caze)
        async with aiofiles.open(out_log, 'w') as outlog, \
                aiofiles.open(out_err, 'w') as errlog:
            async with asyncio.TaskGroup(f'writing {name} log files') as group:
                group.create_task(outlog.write(out))
                group.create_task(errlog.write(err))
        if rc != caze.expected_result:
            out = out.strip()
            err = err.strip()
            msg = f'Test \'{name}\' failed (returned: {rc}, expected: {caze.expected_result}) !'
            if out:
                msg += '\nstdout: ' + out
            if err:
                msg += '\nstderr: ' + err
            raise RuntimeError(msg)

        if caze.expected_output is not None:

            if callable(caze.expected_output):
                expected_output = caze.expected_output(caze)
            else:
                expected_output = caze.expected_output

            if caze.strip_output:
                out = out.strip()

            if caze.normalize_newlines:
                out = out.replace(os.linesep, '\n')

            if isinstance(expected_output, re.Pattern):
                if not expected_output.match(out):
                    out = out.strip()
                    err = err.strip()
                    msg = f'Test \'{name}\' failed (output: {out}, expected: {expected_output}) !'
                    raise RuntimeError(msg)
            elif out != expected_output:
                out = out.strip()
                err = err.strip()
                msg = f'Test \'{name}\' failed (output: {out}, expected: {expected_output.strip()}) !'
                raise RuntimeError(msg)

    async def run_test(self):
        try:
            if len(self.cases) == 0:
                self.error('%s contains no test case', self.name)
                return False
            async with asyncio.TaskGroup(f'running {self.name} tests') as tests:
                for caze in self.cases:
                    tests.create_task(self._run_test(caze))
        except asyncio.ExceptionGroup as errors:
            for err in errors.errors:
                self.error(err)
            return False
        return True

    def __len__(self):
        return len(self.cases)



from random import random
import shutil
from dan.core import asyncio
from dan.core.terminal import TermStream, write as term_write, manager as term_manager

import logging

logging.basicConfig(level=logging.DEBUG)


async def test_status(stream: TermStream, *args, **kwargs):
    await stream.status(*args, **kwargs)
    await asyncio.sleep(0.1)

async def main():
    logger = logging.getLogger('main')
    
    logger.info('start')

    w = shutil.get_terminal_size()[0]

    streams: list[TermStream] = [TermStream(f"output {ii}") for ii in range(5)]

    logger.debug('up all')
    logger.debug('very long debug string, ' * 50)

    streams[0].status('very long status, ' * 50)
    
    await asyncio.sleep(0.1)

    for ii, stream in enumerate(streams):
        stream.status("up")

    await asyncio.sleep(0.1)
    
    logger.debug('up evens')

    for ii, stream in enumerate(streams):
        if ii % 2 == 0:
            stream.status("up even")
        else:
            logger.error('test !' + '.' * w)

    await asyncio.sleep(0.1)
    
    logger.debug('up odds')

    for ii, stream in enumerate(streams):
        if ii % 2 != 0:
            stream.status("up odd")

    await asyncio.sleep(0.5)

    term_write('regular write')

    async def do_stuff(stream: TermStream, **kwargs):
        sub = stream.sub('sub')
        slog = logging.getLogger(stream.name)
        with sub.progress("doing stuff", **kwargs) as bar:
            for n in range(100):
                sub.status(f'item {n}')
                bar()
                slog.error('A' * int(stream.prefix_width * random() * 3))
                await asyncio.sleep(0.025)
        stream.status("done", timeout=1)
        del sub

    logger.debug('testing progresses')

    async with asyncio.TaskGroup() as g:
        for ii, stream in enumerate(streams):
            if ii % 2 == 0:
                g.create_task(do_stuff(stream, total=100))
            else:
                g.create_task(do_stuff(stream))

    logger.debug('testing task group')
    s = streams[3]
    async with s.task_group('task group') as g:
        async def do_stuff(ii):
            await asyncio.sleep(ii * 0.25)
        for ii in range(5):
            g.create_task(do_stuff(ii))

    logger.info('done')

    await asyncio.sleep(1.1)
    term_manager().stop()
    await asyncio.sleep(0.1)


asyncio.run(main())



from typing import Iterable
from dan.core.pathlib import Path
from dan.core import aiofiles
from dan.src.base import SourcesProvider
from dan.core.runners import async_run


class GitSources(SourcesProvider, internal=True):

    url: str = None
    refspec: str = None
    patches: Iterable = list()

    def __init__(self, *args, url=None, refspec=None, patches=None, subdirectory=None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if url is not None:
            self.url = url
        if refspec is not None:
            self.refspec = refspec
        if patches is not None:
            self.patches = patches
        self.sha1 = None
        self.output = self.name
        self.git_dir: Path = self.output / '.git'
        self.subdirectory = subdirectory

    async def __build__(self):
        try:
            self.output.mkdir()            
            await async_run(f'git init -q', logger=self, cwd=self.output)
            await async_run(f'git remote add origin {self.url}', logger=self, cwd=self.output)
            if self.subdirectory is not None:
                # use sparse
                await async_run(f'git config core.sparseCheckout true', logger=self, cwd=self.output)
                async with aiofiles.open(self.git_dir / 'info' / 'sparse-checkout', 'w') as f:
                    await f.writelines([self.subdirectory])

            await async_run(f'git fetch -q --depth 1 origin {self.refspec}', logger=self, cwd=self.output)
            await async_run(f'git checkout -q FETCH_HEAD', logger=self, cwd=self.output)
            
            for patch in self.patches:
                await async_run(f'git am {self.source_path / patch}', logger=self, cwd=self.output)

        except Exception as e:
            await aiofiles.rmtree(self.output)
            raise e




import os
import re

from dan.core import aiofiles, asyncio
from dan.core.pathlib import Path
from dan.core.settings import InstallMode, InstallSettings
from dan.core.target import Target
from dan.core.find import find_file, find_files
from dan.core.version import Version, VersionSpec
from dan.io.repositories import get_packages_path, get_repo_instance
from dan.src.base import SourcesProvider


class PackageBuild(Target, internal=True):

    inherits_version = False
    
    def __init__(self, name, version, repository, package_makefile, *args, spec: VersionSpec = None, **kwargs):
        self.spec = spec
        self.pn = name
        super().__init__(name, *args, version=version, **kwargs)
        self.repo = repository
        self.preload_dependencies.add(self.repo)
        self.package_makefile = package_makefile
        self._build_path = None
        self.toolchain = self.context.get('cxx_target_toolchain')
        self.lock: aiofiles.FileLock = None
        self.__up_to_date = True

    @property
    def is_requirement(self) -> bool:
        return True
    
    @property
    def sources_targets(self):
        targets = []
        for target in self.package_makefile.all_installed:
            for dep in target.preload_dependencies.all:
                if isinstance(dep, SourcesProvider):
                    targets.append(dep)
        return targets
    
    @property
    def target(self):
        if self._target is None:
            self.__resolve()
        return self._target

    async def __initialize__(self):
        source_targets = self.sources_targets
        sources = source_targets[0]
        if self.spec is not None:
            avail_versions = await sources.available_versions()
            if avail_versions is None:
                self.warning(f'unable to get available versions, default one will be used')
            else:
                for version in avail_versions:
                    if self.spec.is_compatible(version):
                        self.debug(f'using version {version} to match {self.spec}')
                        self.version = version
                        break

        packages_path = get_packages_path()
        makefile = self.package_makefile

        # set package version
        version_option = makefile.options.get('version')
        if self.version is None:
            self.version = Version(version_option.value)
        else:
            version_option.value = str(self.version)

        pkgs_root = packages_path / self.toolchain.system / self.toolchain.arch / self.toolchain.build_type.name
        makefile.pkgs_path = pkgs_root / self.name / str(self.version)
        src_path = packages_path / 'src' / self.name / str(self.version)

        self._build_path = makefile.pkgs_path
        self.lock = aiofiles.FileLock(self.build_path / 'build.lock')

        self.install_settings = InstallSettings(self.build_path)
        
        # update package build-path
        makefile.build_path = self.build_path / 'build'
        
        for target in self.package_makefile.all_installed:
            for provided in target.provides:
                pkg_file = find_file(rf'(lib)?{provided}.pc', [self.install_settings.libraries_destination, self.install_settings.data_destination], re.IGNORECASE)
                if pkg_file is None:
                    self.__up_to_date = False
                    break

        for target in self.package_makefile.all_installed:
            for source_target in target.preload_dependencies.all:
                if isinstance(source_target, SourcesProvider):
                    source_target.output = src_path
                    if target.subdirectory is not None:
                        source_target.output /= target.subdirectory

        return await super().__initialize__()
    
    @property
    def up_to_date(self):
        return self.__up_to_date
    
    @property
    def build_path(self) -> Path:
        return self._build_path
    
    async def __build__(self):
        if self.lock.locked:
            self.info('package %s %s is locked, waiting for it to be released...', self.name, self.version)
            # wait for it
            async with self.lock:
                if self.up_to_date:
                    return

        async with self.lock:

            makefile = self.package_makefile
            build_path = makefile.build_path

            async with asyncio.TaskGroup(f'downloading {self.name} sources') as g:
                for target in self.sources_targets:
                    g.create_task(target.build())

            installed = set()
            async def install_target(target):
                if target in installed:
                    return
                installed.add(target)
                for dep in target.target_dependencies:
                    if dep.installed and dep.makefile == self.package_makefile:
                        await install_target(dep)

                self.info(f'installing {target.name}')
                await target.install(self.install_settings, InstallMode.dev)

            async with asyncio.TaskGroup(f'installing {self.name} targets') as g:
                for target in makefile.all_installed:
                    if target in installed:
                        continue
                    g.create_task(install_target(target))

            makefile.cache.ignore()
            del makefile

            os.chdir(self.build_path.parent)

            if not self.toolchain.build_type.is_debug_mode:
                # in debug mode we keep build directory in order to keep debug symbols (might be changed in the future)
                self.debug('cleaning')
                async with asyncio.TaskGroup(f'cleanup {self.name}') as group:
                    group.create_task(aiofiles.rmtree(build_path, force=True))

class ReusePackage(BaseException):
    def __init__(self, pkg):
        self.pkg = pkg

class IoPackage(Target, internal=True):
    
    inherits_version = False

    __all: dict[str, 'IoPackage'] = dict()

    def __init__(self,
                 name: str = None,
                 version: str = None,
                 package: str = None,
                 repository: str = None, **kwargs) -> None:
        self.package = package
        self.repository = repository
        if version is not None:
            self.version = version
        if name is not None:
            self.name = name
        super().__init__(**kwargs)

        self.repo = get_repo_instance(repository, self.makefile)
        self.preload_dependencies.add(self.repo)

    @property
    def is_requirement(self) -> bool:
        return True

    def find(self, name):
        for t in self.package_makefile.all_installed:
            if t.name == name or name in t.provides:
                return t

    @classmethod
    async def instance(cls, name, version, *args, package=None, **kwargs):

        for _, pkg in cls.__all.items():
            if pkg.name == package:
                if version is not None and not version.is_compatible(pkg.version):
                    raise RuntimeError(f'incompatible package version: {pkg.version} {version}')
                return pkg, False
            else:
                target = pkg.find(name)
                if target is not None:
                    if version is not None and not version.is_compatible(pkg.version):
                        raise RuntimeError(f'incompatible package version: {pkg.version} {version}')
                    return pkg, False

        try:
            pkg = IoPackage(name, version, package, *args, **kwargs)
            await pkg.initialize()
            return pkg, True
        except ReusePackage as reuse:
            # concurrent package initialization
            return reuse.pkg, False

    
    async def __initialize__(self):

        self.package_makefile, self.target = self.repo.find(self.name, self.package)
        if self.target is None:
            raise RuntimeError(f'cannot find {self.name} in {self.repo.name}')
        if self.package is None:
            self.package = self.package_makefile.name

        match self.version:
            case str():
                _name, spec = VersionSpec.parse(self.version)
                if _name is not None:
                    name = _name        
                if spec:
                    self.version = spec.version
                    self.spec = spec
                else:
                    self.spec = VersionSpec(Version(self.version), '=')
            case VersionSpec():
                self.spec = self.version
                self.version = self.spec.version
            case Version():
                self.spec = VersionSpec(self.version, '=')
            case None:
                self.spec = None

        if self.package in self.__all:
            other = self.__all[self.package]
            await other.initialize()
            if self.spec is not None and not self.spec.is_compatible(other.version):
                raise RuntimeError(f'duplicate package with incompatible version detected: {self.package} ({self.version} vs {other.version})')
            if self.version is not None and self.version != other.version:
                self.warning(f'using {other.version} instead of {self.version}')
            raise ReusePackage(other)

        self.__all[self.package] = self

        self.pkg_build = PackageBuild(self.package,
                                      self.version,
                                      self.repo,
                                      self.package_makefile,
                                      spec=self.spec,
                                      parent=self)
        self.dependencies.add(self.pkg_build)
        lib_path = Path('pkgs') / 'lib'
        self.pkgconfig_path = lib_path / 'pkgconfig'
        self.cmake_path = lib_path / 'cmake' / self.name
        self.dan_path = lib_path / 'dan'
        
        self.output = self.pkgconfig_path / f'{self.name}.pc'

        return await super().__initialize__()
    
    async def _import_cmake_pkg(self, pkg: Path):
        self.debug('copying %s to %s', pkg, self.build_path / self.cmake_path)
        content = [
            f'set(CMAKE_CURRENT_LIST_FILE "{pkg.absolute().as_posix()}")\n',
            f'set(CMAKE_CURRENT_LIST_DIR "{pkg.parent.absolute().as_posix()}")\n'
        ]
        async with aiofiles.open(pkg) as f:
            content.extend(await f.readlines())
        dest = self.build_path / self.cmake_path / pkg.name
        dest.parent.mkdir(parents=True, exist_ok=True)
        async with aiofiles.open(dest, 'w') as f:
            await f.writelines(content)

    
    async def __build__(self):
        (self.build_path / self.pkgconfig_path).mkdir(exist_ok=True, parents=True)
        (self.build_path / self.dan_path).mkdir(exist_ok=True, parents=True)
        (self.build_path / self.cmake_path).mkdir(exist_ok=True, parents=True)

        async with asyncio.TaskGroup(f'importing {self.name} package') as group:
            for pkg in find_files(r'.+\.pc$', [self.pkg_build.install_settings.libraries_destination / 'pkgconfig']):
                self.debug('copying %s to %s', pkg, self.build_path / self.pkgconfig_path)
                group.create_task(aiofiles.copy(pkg, self.build_path / self.pkgconfig_path))

            for pkg in find_files(r'.+\.pc$', [self.pkg_build.install_settings.data_destination / 'pkgconfig']):
                self.debug('copying %s to %s', pkg, self.build_path / self.pkgconfig_path)
                group.create_task(aiofiles.copy(pkg, self.build_path / self.pkgconfig_path))

            for pkg in find_files(r'.+\.py$', [self.pkg_build.install_settings.data_destination / 'dan']):
                self.debug('copying %s to %s', pkg, self.build_path / self.dan_path)
                group.create_task(aiofiles.copy(pkg, self.build_path / self.dan_path))

            for pkg in find_files(r'.+\.cmake$', [self.pkg_build.install_settings.libraries_destination / 'cmake']):
                self.debug('copying %s to %s', pkg, self.build_path / self.cmake_path)
                group.create_task(self._import_cmake_pkg(pkg))

        if self.output.exists():
            from dan.pkgconfig.package import Data
            data = Data(self.output)
            async with asyncio.TaskGroup(f'importing {self.name} package requirements') as group:
                toolchain = self.context.get('cxx_target_toolchain')
                search_path = get_packages_path() / toolchain.system / toolchain.arch / toolchain.build_type.name
                dest = self.build_path / self.pkgconfig_path
                for pkg in data.requires:
                    pkgconfig_file = find_file(rf'{pkg.name}.pc$', [search_path])
                    # NOTE: find_package will resolve to the build-directory installed pkgconfig, wich will result in a failure
                    # pkg = find_package(pkg.name, spec=pkg.version_spec, search_paths=[search_path], makefile=self.makefile)
                    if pkgconfig_file is not None and not (dest / pkgconfig_file.name).exists():
                        self.debug('copying %s to %s', pkgconfig_file, dest)
                        group.create_task(aiofiles.copy(pkgconfig_file, dest))



import json
import os
from dan.core import asyncio

from dan.core.pm import re_match
from dan.cxx.toolchain import Toolchain

from dan.make import Make
from dan.cxx.targets import CXXObject
from dan.logging import Logging
from dan.core.utils import unique
from dan.core.pathlib import Path


def set_exception_breakpoint(
    *exceptions,
    condition: str = None,
    expression=None,
    notify_on_handled_exceptions=True,
    notify_on_unhandled_exceptions=True,
    notify_on_user_unhandled_exceptions=True,
    notify_on_first_raise_only=False,
):
    try:
        from pydevd import GetGlobalDebugger

        debugger = GetGlobalDebugger()
        for exception in exceptions:
            if isinstance(exception, type):
                exception = exception.__name__
            debugger.add_break_on_exception(
                exception,
                condition,
                expression,
                notify_on_handled_exceptions,
                notify_on_unhandled_exceptions,
                notify_on_user_unhandled_exceptions,
                notify_on_first_raise_only,
            )
    except ImportError:
        print("set_exception_breakpoint: not in vscode environment")


def get_intellisense_mode(toolchain: Toolchain):
    mode = list()
    if toolchain.system is not None:
        if toolchain.system.startswith("msys"):
            mode.append("windows")
        else:
            mode.append(toolchain.system)
    mode.append(toolchain.type)
    mode.append(toolchain.arch)
    return "-".join(mode)


class Code(Logging):
    fullname = "code"

    def __init__(self, make: Make) -> None:
        self.make = make

    def get_test_suites(self, pretty):
        from dan.core.include import MakeFile
        from dan.core.test import Test, Case
        from dan.cxx import Executable

        def make_inner_test_info(test: Test, case: Case):
            basename = test.basename(case)
            out, err = test.outs(case)
            ident = (
                f"{test.fullname}:{case.name}"
                if case.name is not None
                else test.fullname
            )
            info = {
                "type": "test",
                "id": ident,
                "label": basename,
                "debuggable": False,
                "target": test.executable.fullname,
                "out": str(out),
                "err": str(err),
            }
            if isinstance(test.executable, Executable):
                info["debuggable"] = True
                if case.file:
                    info["file"] = str(case.file)
                elif test.file:
                    info["file"] = str(test.file)
                else:
                    info["file"] = str(
                        test.executable.source_path / test.executable.sources[0]
                    )

                if case.lineno:
                    info["line"] = case.lineno
                elif test.lineno:
                    info["line"] = test.lineno

                if test.workingDir:
                    info["workingDirectory"] = str(test.workingDir)
                else:
                    info["workingDirectory"] = str(test.executable.build_path)

                if len(case.args) > 0:
                    info["args"] = [str(a) for a in case.args]

            return info

        def make_test_info(test: Test):
            if len(test) == 0:
                raise RuntimeError(f"Test: {test.name} has not test")
            if len(test) == 1:
                return make_inner_test_info(test, test.cases[0])
            else:
                return {
                    "type": "suite",
                    "id": test.fullname,
                    "label": test.name,
                    "children": [
                        make_inner_test_info(test, case) for case in test.cases
                    ],
                }

        def make_suite_info(mf: MakeFile):
            if len(mf.tests) == 0 and mf.children == 0:
                return None
            children = list()
            for test in mf.tests:
                children.append(make_test_info(test))

            for child in mf.children:
                child_suite = make_suite_info(child)
                if child_suite is not None:
                    children.append(child_suite)

            if len(children) > 0:
                return {
                    "type": "suite",
                    "id": mf.fullname,
                    "label": mf.name,
                    "children": children,
                }

        return json.dumps(
            make_suite_info(self.make.context.root), indent=2 if pretty else None
        )

    async def _init_target(self, target):
        with target.skip_missing_dependencies:
            await target.initialize()

    async def _make_source_configuration(self, source: Path, target: CXXObject):
        # interface:
        #   - includePath: string[]
        #   - defines: string[]
        #   - intelliSenseMode?: "linux-clang-x86" | "linux-clang-x64" | "linux-clang-arm" | "linux-clang-arm64" | "linux-gcc-x86" | "linux-gcc-x64" | "linux-gcc-arm" | "linux-gcc-arm64" | "macos-clang-x86" | "macos-clang-x64" | "macos-clang-arm" | "macos-clang-arm64" | "macos-gcc-x86" | "macos-gcc-x64" | "macos-gcc-arm" | "macos-gcc-arm64" | "windows-clang-x86" | "windows-clang-x64" | "windows-clang-arm" | "windows-clang-arm64" | "windows-gcc-x86" | "windows-gcc-x64" | "windows-gcc-arm" | "windows-gcc-arm64" | "windows-msvc-x86" | "windows-msvc-x64" | "windows-msvc-arm" | "windows-msvc-arm64" | "msvc-x86" | "msvc-x64" | "msvc-arm" | "msvc-arm64" | "gcc-x86" | "gcc-x64" | "gcc-arm" | "gcc-arm64" | "clang-x86" | "clang-x64" | "clang-arm" | "clang-arm64";
        #   - standard?: "c89" | "c99" | "c11" | "c17" | "c++98" | "c++03" | "c++11" | "c++14" | "c++17" | "c++20" | "gnu89" | "gnu99" | "gnu11" | "gnu17" | "gnu++98" | "gnu++03" | "gnu++11" | "gnu++14" | "gnu++17" | "gnu++20";
        #   - forcedInclude?: string[];
        #   - compilerPath?: string;
        #   - compilerArgs?: string[];
        #   - windowsSdkVersion?: string;
        includes = await target.toolchain.get_default_include_paths()
        defines = [
            f"{k}={v}"
            for k, v in (await target.toolchain.get_default_defines()).items()
        ]
        await self._init_target(target)
        for flag in target.private_cxx_flags:
            match re_match(flag):
                case r"[/-]I:?(.+)" as m:
                    includes.append(m[1])
                case r"[/-]D:?(.+)" as m:
                    defines.append(m[1])

        config = {
            "includePath": [os.path.normcase(i) for i in unique(includes)],
            "defines": defines,
            "compilerPath": os.path.normcase(target.toolchain.cxx),
            "intelliSenseMode": get_intellisense_mode(target.toolchain),
            "compilerArgs": target.cxx_flags,
        }
        if target.cpp_std is not None:
            config["standard"] = f"c++{target.cpp_std}"
    
        return {
            "uri": str(source),
            "configuration": config,
        }

    async def get_sources_configuration(self, sources):
        targets_map = await self.make.targets_of(sources)

        async with asyncio.TaskGroup() as g:    
            for source, target in targets_map.items():
                if target:
                    g.create_task(self._make_source_configuration(source, target))
        return json.dumps(g.results())

    async def get_workspace_browse_configuration(self):
        # interface:
        #   - browsePath: string[];
        #   - compilerPath?: string;
        #   - compilerArgs?: string[];
        #   - standard?: see above
        #   - windowsSdkVersion?: string
        from dan.cxx import target_toolchain as toolchain
        from dan.cxx.targets import CXXTarget

        cpp_std = 11
        browse_path = set()
        compiler_args = set()
        cxx_targets = [
            t for t in self.make.root.all_default if isinstance(t, CXXTarget)
        ]
        async with asyncio.TaskGroup("initializing cxx targets") as g:
            for target in cxx_targets:
                g.create_task(self._init_target(target))

        for target in cxx_targets:
            browse_path.update(target.includes.private_raw)
            browse_path.update(target.includes.public_raw)
            compiler_args.update(target.cxx_flags)
            if target.cpp_std is not None and target.cpp_std > cpp_std:
                cpp_std = target.cpp_std

        from dan.pkgconfig.package import get_packages_cache

        for package in get_packages_cache().values():
            compiler_args.update(package.cxx_flags)

        result = {
            "browsePath": [os.path.normcase(p) for p in browse_path],
            "compilerPath": str(toolchain.cxx),
            "compilerArgs": list(compiler_args),
            "standard": f"c++{cpp_std}",
        }
        return json.dumps(result)




import re
from typing import Callable


class ReMatch:
    """Regular expression pattern matchin implementation
    """

    def __init__(self, s: str, fn: Callable[[re.Pattern, str], re.Match]) -> None:
        self._s = s
        self._fn = fn
        self._m = re.Match = None

    __match_args__ = ('_s', '_m')

    def __eq__(self, pattern: str | re.Pattern | tuple[str, int | re.RegexFlag]):
        if isinstance(pattern, str):
            pattern = re.compile(pattern)
        elif isinstance(pattern, tuple):
            pattern = re.compile(*pattern)
        self._m = self._fn(pattern, self._s)
        return self._m is not None
    
    def __str__(self) -> str:
        return self._s

    def __getitem__(
        self,
        group: int | str | tuple[int, ...] | tuple[str, ...]
    ) -> str | tuple[str, ...] | None:
        return self._m[group]

def re_search(string: str) -> ReMatch:
  return ReMatch(string, re.search)


def re_match(string: str) -> ReMatch:
  return ReMatch(string, re.match)


def re_fullmatch(string: str) -> ReMatch:
  return ReMatch(string, re.fullmatch)



import inspect
from dan.core.pathlib import Path
import os


class chdir:
    def __init__(self, path: Path, create=True, strict=False):
        self.path = path
        self.strict = strict
        if create:
            self.path.mkdir(parents=True, exist_ok=True)
        self.prev = None

    def __enter__(self):
        self.prev = Path.cwd()
        os.chdir(self.path)
        return None

    def __exit__(self, *args):
        try:
            os.chdir(self.prev)
        except OSError:
            if self.strict:
                raise


def unique(*seqs):
    seen = set()
    full = list()
    for seq in seqs:
        full.extend(seq)
    return [x for x in full if not (x in seen or seen.add(x))]


def chunks(lst, chunk_size):
    for ii in range(0, len(lst), chunk_size):
        yield lst[ii:ii + chunk_size]


class _ClassPropertyDescriptor(object):

    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)
        return self.fget.__get__(obj, klass)()

    def __set__(self, obj, value):
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func):
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self

def classproperty(func):
    if not isinstance(func, (classmethod, staticmethod)):
        args = inspect.getfullargspec(func).args
        match len(args):
            case 0:
                func = staticmethod(func)
            case 1:
                func = classmethod(func)
            case _:
                raise AttributeError("classproperty can only have 0 or 1 argument (the class object)")

    return _ClassPropertyDescriptor(func)



import aiofiles
from dan.core.pathlib import Path
from dan.core.target import Target, TargetDependencyLike
from dan.core import asyncio
from typing import Callable
import inspect


class generator:
    def __init__(self, output: str, template: str, dependencies: TargetDependencyLike = None, options: dict = None):
        self.output = Path(output)
        self.dependencies = list() if dependencies is None else dependencies
        self.template = template
        self.options = dict() if options is None else options

    def __call__(self, fn: Callable):
        class JinjaGenerator(Target):
            name = self.output.stem
            output = self.output
            template = self.template
            dependencies = [*self.dependencies, self.template]
            options = self.options

            async def __build__(self):
                import jinja2
                arg_spec = inspect.getfullargspec(fn)
                if 'self' in arg_spec.args:
                    data = await asyncio.may_await(fn(self))
                elif not arg_spec.args:
                    data = await asyncio.may_await(fn())
                else:
                    raise RuntimeError(
                        "Only 'self' is allowed as Generator argument")
                self.output.parent.mkdir(parents=True, exist_ok=True)
                env = jinja2.Environment(
                    loader=jinja2.FileSystemLoader(self.source_path))
                template = env.get_template(self.template)
                async with aiofiles.open(self.output, 'w') as out:
                    await out.write(template.render(data))

        # hack the module location (used for Makefile's Targets resolution)
        JinjaGenerator.__module__ = fn.__module__
        return JinjaGenerator



from dan.core import aiofiles
from dan.core.find import find_executable
from dan.core.pathlib import Path
from dan.core.runners import async_run
from dan.cxx.targets import CXXObject
from dan.core.target import Target
from dan.pkgconfig.package import find_package


import functools


class _UIObject(Target, internal=True):

    def __init__(self, ui_file: Path, parent, *args, **kwargs) -> None:
        self.ui_file = Path(ui_file)
        name = 'ui_' + self.ui_file.with_suffix('.h').name
        super().__init__(name, parent, *args, **kwargs)
        self.output = self.ui_file.with_name(name)
        if not self.ui_file.is_absolute():
            self.ui_file = self.source_path / self.ui_file
        self.dependencies.add(self.ui_file)
        

    async def __build__(self):
        p = self.parent
        self.output.parent.mkdir(exist_ok=True, parents=True)
        await async_run([p.uic, '-o', self.output,  self.ui_file], logger=p, log=False, cwd=self.build_path, env=self.toolchain.env)

class _RCCObject(CXXObject, internal=True):

    def __init__(self, resource_file: Path, parent, *args, **kwargs) -> None:
        self.resource_file = Path(resource_file)
        name = 'rcc_' + self.resource_file.with_suffix('.cpp').name
        super().__init__(parent.build_path / self.resource_file.with_name(name), parent, *args, **kwargs)
        if not self.resource_file.is_absolute():
            self.resource_file = self.source_path / self.resource_file
        self.dependencies.add(self.resource_file)

    async def __build__(self):
        p = self.parent
        out, err, rc = await async_run([p.rcc, self.resource_file], logger=p, log=False, cwd=self.build_path, env=self.toolchain.env)
        async with aiofiles.open(self.source, 'w') as f:
            await f.write(out)

        await super().__build__()

class _MocSource(Target, internal=True):

    def __init__(self, header_file: Path, parent, *args, **kwargs) -> None:
        self.header_file = Path(header_file)
        name = 'moc_' + self.header_file.with_suffix('.cpp').name
        source_file = self.header_file.with_name(name)
        if source_file.is_absolute():
            source_file = source_file.relative_to(parent.source_path)
        super().__init__(name, parent, *args, **kwargs)
        self.output = source_file
        self.dependencies.add(self.header_file)

    async def __build__(self):
        p = self.parent
        self.output.parent.mkdir(parents=True, exist_ok=True)
        await async_run([p.moc, '-o', self.output,  self.header_file], logger=p, log=False, cwd=self.build_path, env=self.toolchain.env)

class _MocObject(CXXObject, internal=True):

    def __init__(self, header_file: Path, parent, *args, **kwargs) -> None:
        self.header_file = Path(header_file)
        name = 'moc_' + self.header_file.with_suffix('.cpp').name
        source_file = self.header_file.with_name(name)
        if source_file.is_absolute():
            source_file = source_file.relative_to(parent.source_path)
        super().__init__(parent.build_path / source_file, parent, *args, **kwargs)
        if not self.header_file.is_absolute():
            self.header_file = self.source_path / self.header_file
        self.dependencies.add(self.header_file)

    async def __build__(self):
        p = self.parent
        self.source.parent.mkdir(parents=True, exist_ok=True)
        await async_run([p.moc, '-o', self.source,  self.header_file], logger=p, log=False, cwd=self.build_path, env=self.toolchain.env)
        await super().__build__()
    
class _Wrapper:

    @functools.cached_property
    def search_paths(self):
        qt_core = self.get_dependency(f'Qt{self.qt_major}Core')
        return [qt_core.host_bins if hasattr(qt_core, 'host_bins') else qt_core.bindir, qt_core.prefix] 

    @Target.root_cached_property('qt.moc_executable', Path.as_posix, Path)
    def moc(self):
        return find_executable('moc', paths=self.search_paths, default_paths=False)
     
    @Target.root_cached_property('qt.rcc_executable', Path.as_posix, Path)
    def rcc(self):
        return find_executable('rcc', paths=self.search_paths, default_paths=False)
    
    @Target.root_cached_property('qt.uic_executable', Path.as_posix, Path)
    def uic(self):
        return find_executable('uic', paths=self.search_paths, default_paths=False)
    
    async def __initialize__(self):
        self.includes.private.append(self.build_path)

        extra_include_paths = set()
        for ui_file in self.qt_ui_files:
            uio = _UIObject(ui_file, self)
            self.dependencies.add(uio)
            extra_include_paths.add(uio.output.parent.as_posix())
        
        for resource_file in self.qt_resource_files:
            rcco = _RCCObject(resource_file, self)
            self.objs.append(rcco)
            extra_include_paths.add(rcco.source.parent.as_posix())

        self.includes.extend(extra_include_paths, private=True)

        super()._init_sources()

        prev_objs = list(self.objs)

        self.moc_sources = self.cache.get('moc_sources', list())
        if self.qt_build_moc:
            for file in self.moc_sources:
                self.objs.append(_MocObject(file, self))

        await super().__initialize__()

        if not self.up_to_date:
            if self.qt_build_moc:
                self.objs = prev_objs
            
            self.moc_sources = list()
            from dan.core.find import find_files
            search_paths = set()
            for p in self.includes.all_raw:
                if self.source_path == p or self.source_path in p.parents:
                    search_paths.add(p)
            self.debug('looking for source files to moc in: %s', ', '.join([p.as_posix() for p in search_paths]))
            for file in find_files('.+\.h\w*', search_paths):
                out, err, rc = await async_run([self.moc, file], logger=None, log=False, cwd=self.build_path, no_raise=True, env=self.toolchain.env)
                if rc != 0 or len(out) == 0:
                    self.debug('skipping %s: %s%s', file.name, out, err)
                    continue
                self.moc_sources.append(file)
                if self.qt_build_moc:
                    self.objs.append(_MocObject(file, self))
                else:
                    self.dependencies.add(_MocSource(file, self))
            self.cache['moc_sources'] = self.moc_sources
        elif not self.qt_build_moc:
            self.dependencies.add(_MocSource(file, self))

        self.debug('source files to moc: %s', ', '.join([p.name for p in self.moc_sources]))


    async def __clean__(self):
        await super().__clean__()
        self.cache['moc_sources'] = list()


def wrap(modules: list[str] = None, ui_files: list[str] = None, resource_files: list[str] = None, build_moc=True,  major=6):
    if modules is None:
        modules = ['Widgets']
    if ui_files is None:
        ui_files = []
    if resource_files is None:
        resource_files = []
    modules = set(modules)
    modules.add('Core') # mandatory
    def decorator(cls):
        from dan.core.include import context
        @context.current.wraps(cls)
        class QtWapped(_Wrapper, cls):            
            def __init__(self, *args, **kwargs) -> None:
                super().__init__(*args, **kwargs)
                self.dependencies.update([f'Qt{major}{module}' for module in self.qt_modules], public=True)

            qt_modules = modules
            qt_ui_files = ui_files
            qt_resource_files = resource_files
            qt_major = major
            qt_build_moc = build_moc
        return QtWapped
    return decorator




import os
import platform
from shutil import which
import subprocess
import tempfile

from dan.core.version import Version
import logging

_logger = logging.getLogger('os-info')


class CalledProcessErrorWithStderr(subprocess.CalledProcessError):
    def __str__(self):
        ret = super(CalledProcessErrorWithStderr, self).__str__()
        if self.output:
            ret += "\n" + self.output.decode()
        return ret


def load(path):
    """ Loads a file content """
    with open(path, 'r') as handle:
        return handle.read().decode()


def check_output_runner(cmd, stderr=None):
    # Used to run several utilities, like Pacman detect, AIX version, uname, SCM
    d = tempfile.mkdtemp()
    tmp_file = os.path.join(d, "output")
    try:
        # We don't want stderr to print warnings that will mess the pristine outputs
        stderr = stderr or subprocess.PIPE
        cmd = cmd if isinstance(cmd, str) else subprocess.list2cmdline(cmd)
        command = '{} > "{}"'.format(cmd, tmp_file)
        _logger.info("Calling command: {}".format(command))
        process = subprocess.Popen(command, shell=True, stderr=stderr)
        stdout, stderr = process.communicate()
        _logger.info("Return code: {}".format(int(process.returncode)))

        if process.returncode:
            # Only in case of error, we print also the stderr to know what happened
            raise CalledProcessErrorWithStderr(
                process.returncode, cmd, output=stderr)

        output = load(tmp_file)
        try:
            _logger.info("Output: in file:{}\nstdout: {}\nstderr:{}".format(
                output, stdout, stderr))
        except Exception as exc:
            _logger.error("Error logging command output: {}".format(exc))
        return output
    finally:
        try:
            os.rmdir(d)
        except OSError:
            pass


class OSInfo(object):
    """ Usage:
        (os_info.is_linux) # True/False
        (os_info.is_windows) # True/False
        (os_info.is_macos) # True/False
        (os_info.is_freebsd) # True/False
        (os_info.is_solaris) # True/False

        (os_info.linux_distro)  # debian, ubuntu, fedora, centos...

        (os_info.os_version) # 5.1
        (os_info.os_version_name) # Windows 7, El Capitan

        if os_info.os_version > "10.1":
            pass
        if os_info.os_version == "10.1.0":
            pass
    """

    def __init__(self):
        system = platform.system().lower()
        self.name = system
        self.version = None
        self.version_name = None
        self.is_linux = system == "linux"
        self.linux_distro = None
        self.is_msys = system.startswith(
            "ming") or system.startswith("msys_nt")
        self.is_cygwin = system.startswith("cygwin_nt")
        self.is_windows = system == "windows" or self.is_msys or self.is_cygwin
        self.is_macos = system == "darwin"
        self.is_freebsd = system == "freebsd"
        self.is_solaris = system == "sunos"
        self.is_aix = system == "aix"
        self.is_posix = os.pathsep == ':'

        if self.is_linux:
            self._get_linux_distro_info()
        elif self.is_windows:
            self.version = self._get_win_os_version()
            self.version_name = self._get_win_version_name(self.version)
        elif self.is_macos:
            self.name = "macos"
            self.version = Version(platform.mac_ver()[0])
            self.version_name = self._get_osx_version_name(self.version)
        elif self.is_freebsd:
            self.version = self._get_freebsd_version()
            self.version_name = "freebsd %s" % self.version
        elif self.is_solaris:
            self.version = Version(platform.release())
            self.version_name = self._get_solaris_version_name(
                self.version)
        elif self.is_aix:
            self.version = self._get_aix_version()
            self.version_name = "aix %s" % self.version.minor(fill=False)

    def _get_linux_distro_info(self):
        import distro
        self.linux_distro = distro.id()
        self.version = Version(distro.version())
        version_name = distro.codename()
        self.version_name = version_name if version_name != "n/a" else ""
        if not self.version_name and self.linux_distro == "debian":
            self.version_name = self._get_debian_version_name(self.version)

    @property
    def arch(self):
        match platform.machine():
            case 'i386':
                return 'x86'
            case 'AMD64' | 'x86_64':
                return 'x64'
            case _ as arch:
                return arch

    @property
    def with_apt(self):
        if not self.is_linux:
            return False

        # https://github.com/conan-io/conan/issues/8737 zypper-aptitude can fake it
        if "opensuse" in self.linux_distro or "sles" in self.linux_distro:
            return False

        apt_location = which('apt-get')
        if apt_location:
            # Check if we actually have the official apt package.
            try:
                output = check_output_runner([apt_location, 'moo'])
            except CalledProcessErrorWithStderr:
                return False
            else:
                # Yes, we have mooed today. :-) MOOOOOOOO.
                return True
        else:
            return False

    @property
    def with_yum(self):
        return self.is_linux and self.linux_distro in ("pidora", "fedora", "scientific", "centos",
                                                       "redhat", "rhel", "xenserver", "amazon",
                                                       "oracle", "amzn", "almalinux", "rocky")

    @property
    def with_dnf(self):
        return self.is_linux and self.linux_distro == "fedora" and which('dnf')

    @property
    def with_pacman(self):
        if self.is_linux:
            return self.linux_distro in ["arch", "manjaro"]
        elif self.is_windows and which('uname.exe'):
            uname = check_output_runner(['uname.exe', '-s'])
            return uname.startswith('MSYS_NT') and which('pacman.exe')
        return False

    @property
    def with_zypper(self):
        if not self.is_linux:
            return False
        if "opensuse" in self.linux_distro or "sles" in self.linux_distro:
            return True
        return False

    @staticmethod
    def _get_win_os_version():
        """
        Get's the OS major and minor versions.  Returns a tuple of
        (OS_MAJOR, OS_MINOR).
        """
        import ctypes

        class _OSVERSIONINFOEXW(ctypes.Structure):
            _fields_ = [('dwOSVersionInfoSize', ctypes.c_ulong),
                        ('dwMajorVersion', ctypes.c_ulong),
                        ('dwMinorVersion', ctypes.c_ulong),
                        ('dwBuildNumber', ctypes.c_ulong),
                        ('dwPlatformId', ctypes.c_ulong),
                        ('szCSDVersion', ctypes.c_wchar * 128),
                        ('wServicePackMajor', ctypes.c_ushort),
                        ('wServicePackMinor', ctypes.c_ushort),
                        ('wSuiteMask', ctypes.c_ushort),
                        ('wProductType', ctypes.c_byte),
                        ('wReserved', ctypes.c_byte)]

        os_version = _OSVERSIONINFOEXW()
        os_version.dwOSVersionInfoSize = ctypes.sizeof(os_version)
        if not hasattr(ctypes, "windll"):
            return None
        retcode = ctypes.windll.Ntdll.RtlGetVersion(ctypes.byref(os_version))
        if retcode != 0:
            return None

        return Version("%d.%d" % (os_version.dwMajorVersion, os_version.dwMinorVersion))

    @staticmethod
    def _get_debian_version_name(version):
        if not version:
            return None
        elif version.major == 8:
            return "jessie"
        elif version.major == 7:
            return "wheezy"
        elif version.major == 6:
            return "squeeze"
        elif version.major == 5:
            return "lenny"
        elif version.major == 4:
            return "etch"
        elif version >= Version(3, 1):
            return "sarge"
        elif version >= Version(3, 0):
            return "woody"

    @staticmethod
    def _get_win_version_name(version):
        if not version:
            return None
        elif version.major == 5:
            return "Windows XP"
        elif version >= Version(6, 0):
            return "Windows Vista"
        elif version >= Version(6, 1):
            return "Windows 7"
        elif version >= Version(6, 2):
            return "Windows 8"
        elif version >= Version(6, 3):
            return "Windows 8.1"
        elif version >= Version(10):
            return "Windows 10"

    @staticmethod
    def _get_osx_version_name(version):
        if not version:
            return None
        elif version >= Version(10, 13):
            return "High Sierra"
        elif version >= Version(10, 12):
            return "Sierra"
        elif version >= Version(10, 11):
            return "El Capitan"
        elif version >= Version(10, 10):
            return "Yosemite"
        elif version >= Version(10, 9):
            return "Mavericks"
        elif version >= Version(10, 8):
            return "Mountain Lion"
        elif version >= Version(10, 7):
            return "Lion"
        elif version >= Version(10, 6):
            return "Snow Leopard"
        elif version >= Version(10, 5):
            return "Leopard"
        elif version >= Version(10, 4):
            return "Tiger"
        elif version >= Version(10, 3):
            return "Panther"
        elif version >= Version(10, 2):
            return "Jaguar"
        elif version >= Version(10, 1):
            return "Puma"
        elif version >= Version(10, 0):
            return "Cheetha"

    @staticmethod
    def _get_aix_architecture():
        processor = platform.processor()
        if "powerpc" in processor:
            kernel_bitness = OSInfo()._get_aix_conf("KERNEL_BITMODE")
            if kernel_bitness:
                return "ppc64" if kernel_bitness == "64" else "ppc32"
        elif "rs6000" in processor:
            return "ppc32"

    @staticmethod
    def _get_solaris_architecture():
        # under intel solaris, platform.machine()=='i86pc' so we need to handle
        # it early to suport 64-bit
        processor = platform.processor()
        kernel_bitness, elf = platform.architecture()
        if "sparc" in processor:
            return "sparcv9" if kernel_bitness == "64bit" else "sparc"
        elif "i386" in processor:
            return "x86_64" if kernel_bitness == "64bit" else "x86"

    @staticmethod
    def _get_e2k_architecture():
        return {
            "E1C+": "e2k-v4",  # Elbrus 1C+ and Elbrus 1CK
            "E2C+": "e2k-v2",  # Elbrus 2CM
            "E2C+DSP": "e2k-v2",  # Elbrus 2C+
            "E2C3": "e2k-v6",  # Elbrus 2C3
            "E2S": "e2k-v3",  # Elbrus 2S (aka Elbrus 4C)
            "E8C": "e2k-v4",  # Elbrus 8C and Elbrus 8C1
            "E8C2": "e2k-v5",  # Elbrus 8C2 (aka Elbrus 8CB)
            "E12C": "e2k-v6",  # Elbrus 12C
            "E16C": "e2k-v6",  # Elbrus 16C
            "E32C": "e2k-v7",  # Elbrus 32C
        }.get(platform.processor())

    @staticmethod
    def _get_freebsd_version():
        return platform.release().split("-")[0]

    @staticmethod
    def _get_solaris_version_name(version):
        if not version:
            return None
        elif version >= Version(5, 10):
            return "Solaris 10"
        elif version >= Version(5, 11):
            return "Solaris 11"

    @staticmethod
    def _get_aix_version():
        try:
            return Version(check_output_runner("oslevel").strip())
        except Exception:
            return Version(platform.version(), platform.release())

    @staticmethod
    def uname(options=None):
        options = " %s" % options if options else ""
        if not OSInfo().is_windows:
            raise RuntimeError("Command only for Windows operating system")
        custom_bash_path = OSInfo.bash_path()
        if not custom_bash_path:
            raise RuntimeError("bash is not in the path")

        command = '"%s" -c "uname%s"' % (custom_bash_path, options)
        try:
            # the uname executable is many times located in the same folder as bash.exe
            with environment_append({"PATH": [os.path.dirname(custom_bash_path)]}):
                ret = check_output_runner(command).strip().lower()
                return ret
        except Exception:
            return None

    @staticmethod
    def _get_aix_conf(options=None):
        options = " %s" % options if options else ""
        if not OSInfo().is_aix:
            raise RuntimeError("Command only for AIX operating system")

        try:
            ret = check_output_runner("getconf%s" % options).strip()
            return ret
        except Exception:
            return None

    @staticmethod
    def _detect_windows_subsystem():
        from conans.client.tools.win import CYGWIN, MSYS2, MSYS, WSL
        if OSInfo().is_linux:
            try:
                # https://github.com/Microsoft/WSL/issues/423#issuecomment-221627364
                with open("/proc/sys/kernel/osrelease") as f:
                    return WSL if f.read().endswith("Microsoft") else None
            except IOError:
                return None
        try:
            output = OSInfo.uname()
        except Exception:
            return None
        if not output:
            return None
        if "cygwin" in output:
            return CYGWIN
        elif "msys" in output or "mingw" in output:
            version = OSInfo.uname("-r").split('.')
            if version and version[0].isdigit():
                major = int(version[0])
                if major == 1:
                    return MSYS
                elif major >= 2:
                    return MSYS2
            return None
        elif "linux" in output:
            return WSL
        else:
            return None
    def __str__(self) -> str:
        return f'{self.linux_distro or self.name} - {self.version_name} ({self.version})'


info = OSInfo()



from pathlib import Path
from dan.core.include import MakeFileError
from tests import PyMakeBaseTest

base_path = Path(__file__).parent / 'errors' / 'python'

class PythonErrors(PyMakeBaseTest):
    def __init__(self, methodName: str = None) -> None:
        super().__init__(methodName=methodName, source_path=base_path)

    async def test_import_error(self):
        async with self.section('import_error', targets=['InvalidSyntax'], subdir='import_error', clean=True, diags=True, init=False) as make:
            try:
                await make.initialize()
                self.fail('MakeFileError should have been raised')
            except MakeFileError:
                diags = make.diagnostics
                self.assertEqual(len(diags), 1)
                filename = str(self.source_path / 'import_error' / 'dan-build.py')
                self.assertTrue(filename in diags)
                self.assertEqual(len(diags[filename]), 1)
                self.assertEqual(diags[filename][0].message, "No module named 'not_a_module'")

    async def test_all_errors(self):
        async with self.section('import_error', targets=['InvalidSyntax'], clean=True, diags=True, init=False) as make:
            try:
                await make.initialize()
                self.fail('MakeFileError should have been raised')
            except MakeFileError:
                diags = make.diagnostics
                self.assertEqual(len(diags), 2)



from functools import cached_property
from dan.core import aiofiles, diagnostics as diag
from dan.core.pm import re_match
from dan.core.settings import BuildType
from dan.core.utils import unique
from dan.cxx.toolchain import CommandArgsList, Toolchain, Path, FileDependency, CppStd
from dan.cxx import auto_fpic
from dan.core.runners import sync_run

import typing as t

cxx_extensions = ['.cpp', '.cxx', '.C', '.cc']
c_extensions = ['.c']


class UnixToolchain(Toolchain):
    def __init__(self, data, tools, *args, **kwargs):
        Toolchain.__init__(self, data, tools, *args, **kwargs)
        self.cc = Path(data['cc'])
        self.cxx = Path(data['cxx'])
        self.ar = data['ar'] if 'ar' in data else tools['ar']
        self.ranlib = data['ranlib'] if 'ranlib' in data else tools['ranlib']
        # self.as_ = data['as'] if 'as' in data else tools['as']
        # self.strip = data['env']['STRIP'] if 'env' in data and 'STRIP' in data['env'] else tools['strip']
        self.env = data['env'] if 'env' in data else None
        self.debug('cxx compiler is %s %s (%s)',
                   self.type, self.version, self.cc)
        self.debug('cxx compiler is %s %s (%s)',
                   self.type, self.version, self.cxx)

    def get_optimization_flags(self, build_type):
        flags = list()
        if build_type is None:
            build_type = self.build_type
        match build_type:
            case BuildType.debug:
                flags.extend(('-g', ))
            case BuildType.release:
                flags.extend(('-O3', '-DNDEBUG'))
            case BuildType.release_min_size:
                flags.extend(('-Os', '-DNDEBUG'))
            case BuildType.release_debug_infos:
                flags.extend(('-O2', '-g', '-DNDEBUG'))
        return flags

    @cached_property
    def default_cflags(self):
        flags = list()
        if self.env:
            if 'SYSROOT' in self.env:
                flags.append(f'--sysroot={self.env["SYSROOT"]}')
            if 'CFLAGS' in self.env:
                flags.extend(self.env["CFLAGS"].strip().split(' '))
        return unique(flags)

    @cached_property
    def default_cxxflags(self):
        flags = [*self.settings.cxx_flags]
        if self.env:
            if 'CXXFLAGS' in self.env:
                flags.extend(self.env["CXXFLAGS"].strip().split(' '))
        return unique(flags)

    @cached_property
    def default_ldflags(self):
        flags = list()
        if self.env:
            if 'LDFLAGS' in self.env:
                flags.extend(self.env["LDFLAGS"].strip().split(' '))
        return unique(flags)

    def has_cxx_compile_options(self, *opts) -> bool:
        _, err, _ = sync_run([self.cxx, *opts], no_raise=True)
        return err.splitlines()[0].find('no input files') >= 0

    def make_include_options(self, include_paths: set[Path]) -> list[str]:
        return unique([f'-I{p}' for p in include_paths])

    def make_libpath_options(self, libraries: set[Path | str]) -> list[str]:
        opts = list()
        if self.rpath:
            opts.append(f'-Wl,-rpath,{self.rpath}')

        for lib in libraries:
            if isinstance(lib, Path):
                opts.append(f'-L{lib.parent}')
                if not self.rpath:
                    opts.append(f'-Wl,-rpath,{lib.parent}')
        return opts

    def make_link_options(self, libraries: set[Path | str]) -> list[str]:
        opts = list()
        for lib in libraries:
            if isinstance(lib, Path):
                opts.append(f'-l{lib.stem.removeprefix("lib")}')
            else:
                assert isinstance(lib, str)
                opts.append(f'-l{lib}')
        return opts

    def make_compile_definitions(self, definitions: set[str]) -> list[str]:
        return unique([f'-D{d}' for d in definitions])

    def make_compile_options(self, options: set[str]) -> list[str]:
        result = list()
        for o in options:
            match o:
                case CppStd():
                    result.append(f'-std=c++{o.stdver}')
                case _:
                    result.append(o)
        return result

    def make_library_name(self, basename: str, shared: bool) -> str:
        if not shared:
            return f'lib{basename}.a'
        elif self.system.is_windows:
            return f'lib{basename}.dll'
        else:
            return f'lib{basename}.so'

    def make_executable_name(self, basename: str) -> str:
        return f'{basename}.exe' if self.system.is_windows else basename

    def get_base_compile_args(self, sourcefile: Path, build_type) -> list[str]:
        match sourcefile.suffix:
            case _ if sourcefile.suffix in cxx_extensions:
                return [self.cxx, *self.default_cxxflags, *self.get_optimization_flags(build_type), *self.default_cflags]
            case _ if sourcefile.suffix in c_extensions:
                return [self.cc, *self.get_optimization_flags(build_type), *self.default_cflags]
            case _:
                raise RuntimeError(
                    f'Unhandled source file extention: {sourcefile.suffix}')

    async def scan_dependencies(self, sourcefile: Path, output: Path, options: set[str]) -> set[FileDependency]:
        deps_path = output.with_suffix(".o.d")
        deps = list()
        if deps_path.exists():
            async with aiofiles.open(deps_path, 'r') as f:
                for dep in await f.readlines():
                    dep = dep.strip()
                    if dep.endswith('\\'):
                        dep = dep[:-2]
                    deps.append(dep.strip())
                _obj = deps.pop(0)
                if len(deps) > 0:
                    _src = deps.pop(0)
        return set(deps)

    def compile_generated_files(self, output: Path) -> set[Path]:
        return {output.with_suffix(output.suffix + '.d')}

    @property
    def cxxmodules_flags(self) -> list[str]:
        return ['-std=c++20', '-fmodules-ts']

    def make_compile_commands(self, sourcefile: Path, output: Path, options: set[str], build_type=None) -> CommandArgsList:
        args = self.get_base_compile_args(sourcefile, build_type)
        args.extend([*self.compile_options, *options, '-MD', '-MT', str(output),
                    '-MF', f'{output}.d', '-o', str(output), '-c', str(sourcefile)])
        if auto_fpic:
            args.insert(1, '-fPIC')
        return [args]

    def make_link_commands(self, objects: set[Path], output: Path, options: list[str]) -> CommandArgsList:
        args = [self.cxx, *objects, '-o', str(output), *unique(
            self.default_ldflags, self.default_cflags, self.default_cxxflags, self.link_options, options)]
        commands = [args]
        if self._build_type in [BuildType.release, BuildType.release_min_size]:
            commands.append([self.strip, output])
        return commands

    def make_static_lib_commands(self, objects: set[Path], output: Path, options: list[str]) -> CommandArgsList:
        return [
            [self.ar, 'cr', output, *objects],  # *options],
            [self.ranlib, output],
        ]

    def make_shared_lib_commands(self, objects: set[Path], output: Path, options: list[str]) -> tuple[Path, CommandArgsList]:
        args = [self.cxx, '-shared', *objects, *
                unique(self.default_ldflags, options), '-o', output]
        commands = [args]
        if self._build_type in [BuildType.release, BuildType.release_min_size]:
            commands.append([self.strip, output])
        return commands

    async def get_default_include_paths(self, lang='c++') -> list[Path]:
        cache_key = f'default_{lang}_includes'
        includes = self.cache.get(cache_key, None)
        if includes is None:
            args = [self.cc, '-x', lang, '-E', '-Wp,-v', '-']
            out, err, rc = await self.run(f'get default {lang} includes', None, args, quiet=True, log=False, input='')
            if rc != 0:
                self.error('failed to get default %s includes: %s', lang, err)
                return []
            includes = []
            for line in [*out.splitlines(), *err.splitlines()]:
                match re_match(line):
                    case r'^ (.+)$' as m:
                        includes.append(str(Path(m[1]).resolve()))
            self.cache[cache_key] = includes
        return includes

    async def _gen_gcc_compile_diags(self, lines) -> t.Iterable[diag.Diagnostic]:
        _from = list()
        prev: diag.Diagnostic | diag.RelatedInformation = None
        prev_diag: diag.Diagnostic = None
        async for line in lines:
            match re_match(line):
                case r'\s+?\|\s(\s+)?(\^~+)' as m:
                    if prev is not None:
                        if isinstance(prev, diag.Diagnostic):
                            rng = prev.range
                        else:
                            rng = prev.location.range
                        rng.start.character = len(m[1]) if m[1] else 0
                        rng.end.character = rng.start.character + len(m[2])
                case r'((?:.+)from (.+)):(\d+)[,:]' as m:
                    message = m[1]
                    if message.startswith('In file included'):
                        _from.clear()
                    filename = m[2]
                    lineno = int(m[3]) - 1
                    prev = info = diag.RelatedInformation(
                        location=diag.Location(diag.Uri(filename),
                                               range=diag.Range(start=diag.Position(lineno), end=diag.Position(lineno))),
                        message=message)
                    _from.append(info)
                case r'(.+?): In instantiation of \'(.+)\'' as m:
                    _from.clear()
                case r'(.+?):(\d+):(?:(\d+):)?\s+(required from\s.+)' as m:
                    filename = m[1]
                    lineno = int(m[2]) - 1
                    character = int(m[3]) - 1 if m[3] else 0
                    message = m[4]
                    prev = info = diag.RelatedInformation(
                        location=diag.Location(diag.Uri(filename),
                                               range=diag.Range(start=diag.Position(lineno, character), end=diag.Position(lineno, character))),
                        message=message)
                    _from.append(info)
                case r'(.+?):(\d+):(?:(\d+):)?\s(note):\s(.+)$' as m:
                    filename = m[1]
                    character = int(m[3]) if m[3] else 0
                    lineno = int(m[2]) - 1
                    message = m[5]
                    if prev_diag is None:
                        self.warning(
                            'diagnostics: a note is expected to append after a diagnositc')
                    else:
                        info = diag.RelatedInformation(
                            location=diag.Location(diag.Uri(filename),
                                                   range=diag.Range(start=diag.Position(lineno, character), end=diag.Position(lineno, character))),
                            message=message)
                        prev_diag.related_information.insert(0, info)
                        prev = info
                case r'(.+?):(\d+):(?:(\d+):)?\s(?:fatal )?(error|warning):\s(.+)$' as m:
                    character = int(m[3]) if m[3] else 0
                    lineno = int(m[2]) - 1
                    message = m[5]
                    prev_diag = prev = diag.Diagnostic(
                        message=message,
                        range=diag.Range(start=diag.Position(
                            line=lineno, character=character), end=diag.Position(line=lineno)),
                        severity=diag.Severity[m[4].upper()],
                        source=self.type,
                        filename=m[1],
                        related_information=list(_from)
                    )
                    yield prev_diag

    async def _handle_compile_output(self, lines) -> t.Iterable[diag.Diagnostic]:
        match self.type:
            case 'gcc' | 'clang':
                async for d in self._gen_gcc_compile_diags(lines):
                    yield d
            case _:
                raise NotImplementedError(
                    f'handle_compile_output errors not implemented for {self.type}')

    async def _gen_ld_link_diags(self, lines) -> t.Iterable[diag.Diagnostic]:
        function = None
        object = None
        async for line in lines:
            match re_match(line):
                case r'(.+): in function `(.+)\':$' as m:
                    object = m[1]
                    function = m[2]
                case r'(?:.+: )?(?:(.+):)?\((.+)\+(.+)\): (.+)$' as m:
                    # link error may not be associated to a source file,
                    # in which case the associated file is the object
                    filename = m[1] or object
                    section = m[2]
                    section_offset = int(m[3], 0)
                    message = m[4].strip()
                    yield diag.Diagnostic(
                        message=message,
                        source=self.type
                    )
                case r'(?:.+?: )?(.+?):(\d+): (undefined reference to.+)$' as m:
                    filename = m[1]
                    line = int(m[2])
                    message = m[3].strip()
                    yield diag.Diagnostic(
                        message=message,
                        filename=filename,
                        range=diag.Range(start=diag.Position(line=line)),
                        source=self.type
                    )
                case _:
                    self._logger.debug('unhandled line: %s', line)

    async def _handle_link_output(self, lines) -> t.Iterable[diag.Diagnostic]:
        match self.type:
            case 'gcc' | 'clang':
                async for d in self._gen_ld_link_diags(lines):
                    yield d
            case _:
                raise NotImplementedError(
                    f'handle_link_output not implemented for {self.type}')



import os
import re
import typing as t

from collections.abc import Iterable
from enum import Enum
from functools import cached_property

from dan.core.pathlib import Path
from dan.core import cache
from dan.core.target import Target, Installer, InstallMode
from dan.core.utils import chunks, unique
from dan.core.runners import async_run
from dan.core import asyncio
from dan.cxx.toolchain import CompilationFailure, LibraryList, LinkageFailure, Toolchain, CppStd, BuildType
from dan.core.cache import cached_property as dan_cached

class CXXObject(Target, internal=True):
    def __init__(self, source:Path, parent: 'CXXTarget', root: Path = None) -> None:
        if source.is_absolute():
            if root is None:
                root = parent.build_path
            name = '-'.join(source.relative_to(root).with_suffix(f'').parts)
        else:
            name = '-'.join(source.with_suffix(f'').parts)
        super().__init__(name, parent=parent, default=False)
        self.parent = parent
        self.source = source
        self.toolchain: Toolchain = self.context.get('cxx_target_toolchain')
        obj_fname = source.with_suffix('.obj' if self.toolchain.type == 'msvc' else '.o')
        if source.is_absolute():
            if source.parent.is_relative_to(self.parent.source_path):
                rpath = source.parent.relative_to(self.parent.source_path)
                self.output = self.build_path / rpath / obj_fname.name
            elif obj_fname.is_relative_to(self.build_path):
                self.output = obj_fname
            else:
                self.output = self.build_path / obj_fname.name
        else:
            self.output = self.build_path / obj_fname
        self.__dirty = False

    @property
    def build_type(self):
        return self.parent.build_type

    @property
    def cxx_flags(self):
        return self.parent.cxx_flags

    @property
    def private_cxx_flags(self):
        return self.parent.private_cxx_flags
    
    @property
    def includes(self):
        return self.parent.includes

    @property
    def compile_definitions(self):
        return self.parent.compile_definitions
    
    @dan_cached()
    def deps(self): ...

    @dan_cached()
    def compile_args(self): ...

    async def __initialize__(self):
        await self.parent.preload()

        deps = self.deps
        if deps is not None:
            self.dependencies.update(deps)

        self.dependencies.add(self.source)

        self.other_generated_files.update(
            self.toolchain.compile_generated_files(self.output))

        previous_args = self.compile_args
        if previous_args is not None:
            args = self.toolchain.make_compile_commands(
                self.source_path / self.source, self.output, self.private_cxx_flags, self.build_type)[0]
            args = [str(arg) for arg in args]
            if sorted(args) != sorted(previous_args):
                self.__dirty = True
        else:
            self.__dirty = True

    @cached_property
    def up_to_date(self):
        if self.__dirty:
            return False
        return super().up_to_date

    async def __build__(self):
        self.info('generating %s...', self.output.name)
        try:
            self.output.parent.mkdir(parents=True, exist_ok=True)
            commands, diags = await self.toolchain.compile(self.source_path / self.source, self.output, self.private_cxx_flags, self.build_type)
            self.parent.diagnostics.insert(diags, str(self.source))
        except CompilationFailure as err:
            self.parent.diagnostics.insert(err.diags, str(self.source))
            err.target = self
            raise
        self.compile_args = [str(a) for a in commands[0]]
        self.debug('scanning dependencies of %s', self.source.name)
        deps = await self.toolchain.scan_dependencies(self.source_path / self.source, self.output, self.private_cxx_flags)
        deps = [d for d in deps
                if self.makefile.root.source_path in Path(d).parents
                or self.build_path in Path(d).parents]
        self.deps = deps


class OptionSet:
    def __init__(self, parent: 'CXXTarget',
                 name: str,
                 public: list | set = set(),
                 private: list | set = set(),
                 transform_out: t.Callable[[t.Any], t.Any] = None,
                 transform_in: t.Callable[[t.Any], t.Any] = None) -> None:
        self._parent = parent
        self._name = name
        self._transform_out = transform_out or self.__nop_transform
        self._transform_in = transform_in or self.__nop_transform
        self._public = list()
        self._private = list()
        self.add(*public, public=True)
        self.add(*private, public=False)

    @staticmethod
    def __nop_transform(x):
        return x

    @property
    def private(self) -> list:
        return unique(self._transform_out([self._transform_in(p) for p in self._private]))

    @property
    def public(self) -> list:
        items: list = self._transform_out([self._transform_in(p) for p in self._public])
        for dep in self._parent._recursive_dependencies((CXXTarget)):
            opts = getattr(dep, self._name)
            items.extend(opts._transform_out([opts._transform_in(p) for p in opts._public]))
        return unique(items)

    @property
    def all(self) -> list:
        items = list()
        items.extend(self.private)
        items.extend(self.public)
        return unique(items)

    @property
    def private_raw(self) -> list:
        return [self._transform_in(p) for p in self._private]

    @property
    def public_raw(self) -> list:
        return [self._transform_in(p) for p in self._public]
    
    @property
    def all_raw(self) -> list:
        return [*self.private_raw, *self.public_raw]

    def add(self, *values, public=False):
        if public:
            for value in values:
                if not value in self._public:
                    self._public.append(value)
        else:
            for value in values:
                if not value in self._private:
                    self._private.append(value)

    def update(self, values: 'OptionSet', private=False):
        self._public = values._public
        if private:
            self._private = values._private

    def extend(self, values: t.Iterable, private=False):
        if private:
            self._private.extend(values)
        else:
            self._public.extend(values)


class CXXTarget(Target, internal=True):
    public_includes: set[str] = set()
    private_includes: set[str] = set()

    public_compile_options: set[str] = set()
    private_compile_options: set[str] = set()
    
    public_compile_definitions: set[str] = set()
    private_compile_definitions: set[str] = set()

    public_lib_paths: set[str] = set()
    private_lib_paths: set[str] = set()

    public_link_libraries: set[str] = set()
    private_link_libraries: set[str] = set()

    public_link_options: set[str] = set()
    private_link_options: set[str] = set()

    build_type: BuildType = None

    __cpp_std: int|str = None

    def __make_src_path(self, path):
        if not isinstance(path, Path):
            path = Path(path)
        return path if path.is_absolute() else self.source_path / path

    def __init__(self,
                 *args,
                 **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.toolchain : Toolchain = self.context.get('cxx_target_toolchain')

        self.includes = OptionSet(self, 'includes',
                                  self.public_includes, self.private_includes,
                                  transform_out=self.toolchain.make_include_options,
                                  transform_in=self.__make_src_path)

        self.compile_options = OptionSet(self, 'compile_options',
                                         self.public_compile_options, self.private_compile_options,
                                         transform_out=self.toolchain.make_compile_options)

        self.link_libraries = OptionSet(self, 'link_libraries',
                                        self.public_link_libraries, self.private_link_libraries,
                                        transform_out=self.toolchain.make_link_options)
        
        self.library_paths = OptionSet(self, 'library_paths',
                                        self.public_lib_paths, self.private_lib_paths,
                                        transform_out=self.toolchain.make_libpath_options)

        self.compile_definitions = OptionSet(self, 'compile_definitions',
                                             self.public_compile_definitions, self.private_compile_definitions,
                                             transform_out=self.toolchain.make_compile_definitions)

        self.link_options = OptionSet(self, 'link_options',
                                      self.public_link_options, self.private_link_options)

    @property
    def cpp_std(self):
        if self.__cpp_std is None:
            self.__cpp_std = self.makefile.get_attribute('cpp_std', recursive=True)
            if self.__cpp_std is None:
                self.__cpp_std = -1
        if self.__cpp_std == -1:
            return None
        return self.__cpp_std
    
    @cpp_std.setter
    def cpp_std(self, value):
        self.__cpp_std = value

    @property
    def cxx_dependencies(self) -> list['CXXTarget']:
        return [dep for dep in self.dependencies.all if isinstance(dep, CXXTarget)]

    @property
    def library_dependencies(self) -> list['Library']:
        return [dep for dep in self.dependencies.all if isinstance(dep, Library)]

    @cached_property
    def shared_dependencies_path(self):
        paths = []
        for lib in [d for d in self.dependencies.all if isinstance(d, Library) and d.shared]:
            paths.append(lib.build_path.as_posix())
        for target in self.cxx_dependencies:
            paths.extend(target.shared_dependencies_path)
        return paths

    @cached_property
    def lib_paths(self) -> list[str]:
        tmp = set()
        for dep in self.cxx_dependencies:
            tmp.update(dep.lib_paths)
        tmp.update(self.library_paths.public)
        # # TODO move create private_libs()
        tmp.update(self.library_paths.private)
        return list(sorted(tmp))

    @cached_property
    def libs(self) -> LibraryList:
        tmp = LibraryList()
        for dep in reversed(self.cxx_dependencies):
            tmp.extend(dep.libs)
        tmp.extend(self.link_libraries.public)
        # # TODO move create private_libs()
        tmp.extend(self.link_libraries.private)
        return tmp
    
    @property
    def build_type(self):
        return self.toolchain.build_type

    @cached_property
    def cxx_flags(self):
        flags = self.includes.public
        flags.extend(self.compile_options.public)
        flags.extend(self.compile_definitions.public)
        for dep in self.cxx_dependencies:
            flags.extend(dep.cxx_flags)
        return unique(flags)

    @cached_property
    def private_cxx_flags(self):
        flags = []
        cpp_std = self.cpp_std
        if cpp_std is not None:
            flags.extend(self.toolchain.make_compile_options([cpp_std if isinstance(cpp_std, CppStd) else CppStd(cpp_std)]))
        flags.extend(self.includes.private)
        flags.extend(self.cxx_flags)
        flags.extend(self.compile_options.private)
        flags.extend(self.compile_definitions.private)
        return unique(flags)
    
    async def __install__(self, installer: Installer):
        if installer.mode == InstallMode.portable:
            
            exclude_paths = []
            if self.toolchain.system.is_windows:
                exclude_paths.append(Path(os.getenv('SYSTEMROOT')))
            def check(p: Path):
                for e in exclude_paths:
                    if e in p.parents:
                        return False
                return True
            from dan.cxx.ldd import get_runtime_dependencies
            async with asyncio.TaskGroup(f'{self.name} runtime dependencies installation') as g:
                for dep, path in await get_runtime_dependencies(self):
                    if path is not None: # not found ?
                        path = Path(path)
                        if check(path):
                            g.create_task(installer.install_bin(path))
        await super().__install__(installer)

StrOrPath = str|Path
StrOrPathIterable = Iterable[StrOrPath]

class CXXObjectsTarget(CXXTarget, internal=True):
    sources: StrOrPathIterable|t.Callable[[], StrOrPathIterable] = set()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.objs: list[CXXObject] = list()

    @cache.once_method
    def _init_sources(self):
        if callable(self.sources):
            self.sources = list(self.sources())
        if not isinstance(self.sources, Iterable):
            assert callable(
                self.sources), f'{self.name} sources parameter should be an iterable or a callable returning an iterable'
        sources = list()
        if self.source_path != self.makefile.source_path:
            self.sources = [self.source_path / source for source in self.sources]
        if self.sources:
            source_root = Path(os.path.commonpath(self.sources))
            for source in self.sources:
                source = Path(source)
                if source.is_absolute():
                    root = source_root
                    if root.is_file():
                        root = root.parent
                else:
                    root = self.source_path
                sources.append(source)
                self.objs.append(
                    CXXObject(Path(source), self, root=root))
            self.sources = sources
            

    @property
    def file_dependencies(self):
        return unique(super().file_dependencies, *[o.file_dependencies for o in self.objs])
    
    @cached_property
    def up_to_date(self):
        for obj in self.objs:
            if not obj.up_to_date:
                return False
        return super().up_to_date

    @property
    def headers(self):
        return [f for f in self.file_dependencies if f.suffix.startswith('.h')]

    async def __initialize__(self):
        self._init_sources()
        async with asyncio.TaskGroup(f'initializing {self.name}\'s objects') as group:
            for obj in self.objs:
                group.create_task(obj.initialize())
                # self.load_dependency(obj)

    async def __build__(self):
        # compile objects
        async with self.task_group(f'building {self.name}\'s objects') as group:
            for dep in self.objs:
                group.create_task(dep.build())

    async def __clean__(self):
        async with asyncio.TaskGroup(f'cleaning {self.name}\'s objects') as group:
            for dep in self.objs:
                group.create_task(dep.clean())
        return await super().__clean__()


class LibraryType(str, Enum):
    AUTO = 'auto'
    STATIC = 'static'
    SHARED = 'shared'
    INTERFACE = 'interface'


class Library(CXXObjectsTarget, internal=True):

    header_match = r'.+'
    library_type: LibraryType = LibraryType.AUTO

    @property
    def static(self) -> bool:
        return self.library_type == LibraryType.STATIC

    @property
    def shared(self) -> bool:
        return self.library_type == LibraryType.SHARED

    @property
    def interface(self) -> bool:
        return self.library_type == LibraryType.INTERFACE

    @cached_property
    def lib_paths(self) -> list[str]:
        tmp = super().lib_paths
        if not self.interface:
            tmp.extend(self.toolchain.make_libpath_options([self.output]))
        return list(sorted(tmp))
    
    @property
    def libs(self) -> list[str]:
        if not self.interface:
            libs = LibraryList()
            libs.extend(self.toolchain.make_link_options([self.output]))
            libs.extend(super().libs)
        else:
            libs = super().libs
        
        return libs
    
    def __make_link_options(self):
        return [*self.lib_paths, *self.libs, *self.link_options.public, *self.link_options.private]


    async def __initialize__(self):
        self._init_sources()

        if self.library_type == LibraryType.AUTO:
            if len(self.sources) == 0:
                self.library_type = LibraryType.INTERFACE
            else:
                from dan.core.settings import DefaultLibraryType
                if self.toolchain.settings.default_library_type == DefaultLibraryType.static:
                    self.library_type = LibraryType.STATIC
                else:
                    self.library_type = LibraryType.SHARED

        from .msvc_toolchain import MSVCToolchain
        if self.shared and isinstance(self.toolchain, MSVCToolchain):
            self.compile_definitions.add(f'{self.name.upper()}_EXPORT=1')

        if self.library_type != LibraryType.INTERFACE:
            self.output = self.toolchain.make_library_name(self.name, self.shared)
        else:
            self.output = f"lib{self.name}.stamp"
        await super().__initialize__()

        previous_args = self.cache.get('generate_args')
        generate = None
        match self.library_type:
            case LibraryType.STATIC:
                generate = self.toolchain.static_lib
            case LibraryType.SHARED:
                generate = self.toolchain.shared_lib
        if generate is not None:
            if previous_args and \
                    previous_args != await generate(
                        [obj.routput for obj in self.objs], self.output, self.__make_link_options(), dry_run=True):
                self.__dirty = True
            else:
                self.__dirty = False
        else:
            self.__dirty = False

    @cached_property
    def up_to_date(self):
        if self.__dirty:
            return False
        return super().up_to_date

    async def __build__(self):
        await super().__build__()

        self.info(
            'creating %s library %s...', self.library_type.name.lower(), self.output.name)

        if self.static:
            await self.toolchain.static_lib([obj.routput for obj in self.objs], self.output, self.__make_link_options())
        elif self.shared:
            await self.toolchain.shared_lib([obj.routput for obj in self.objs], self.output, self.__make_link_options())
            from .msvc_toolchain import MSVCToolchain
            if isinstance(self.toolchain, MSVCToolchain):
                self.compile_definitions.add(
                    f'{self.name.upper()}_IMPORT=1', public=True)
        else:
            assert self.interface
            self.output.touch()

        self.debug('done')
    
    def __install_headers__(self, installer: Installer) -> list:
        tasks = list()
        header_expr = re.compile(self.header_match)
        for public_include_dir in self.includes.public_raw:
            headers = public_include_dir.rglob('*.h*')
            for header in headers:
                if header_expr.match(header.as_posix()):
                    subdirs = header.relative_to(public_include_dir).parent
                    tasks.append(installer.install_header(header, subdirs))
        return tasks
        


    async def __install__(self, installer: Installer):

        tasks = list()

        if installer.settings.create_pkg_config:
            from dan.pkgconfig.package import create_pkg_config
            tasks.append(create_pkg_config(self, installer.settings))

        if self.shared:
            tasks.append(installer.install_shared_library(self.output))
        elif self.static:
            tasks.append(installer.install_static_library(self.output))

        if installer.dev:
            tasks.extend(self.__install_headers__(installer))

            # TODO: how to handle debug symbols ? check where debug it is usually installed and do the same
            # for obj in self.objs:
            #     for dbg_file in self.toolchain.debug_files(obj.output):
            #         tasks.append(do_install(dbg_file, settings.libraries_destination / dbg_file.name))

        tasks.insert(0, super().__install__(installer))

        for tchunk in chunks(tasks, 100):
            await asyncio.gather(*tchunk)


class Module(CXXObjectsTarget, internal=True):
    def __init__(self, name: str, sources: list[str], *args, **kwargs):
        super().__init__(name, sources, *args, **kwargs)

    @property
    def cxx_flags(self):
        return {*self.toolchain.cxxmodules_flags, *super().cxx_flags}

    async def __build__(self):
        return await super().__build__()

class Executable(CXXObjectsTarget, internal=True):

    installed = True
    subsystem: str = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.subsystem is None:
            self.subsystem = 'console'

        self.output = self.toolchain.make_executable_name(self.name)
        self.__dirty = False

    def _make_link_options(self):
        subsystem_opt = []
        if self.toolchain.type == 'msvc':
            subsystem_opt.append(f'/subsystem:{self.subsystem}')
        return [*subsystem_opt, *self.lib_paths, *self.libs, *self.link_options.public, *self.link_options.private]

    @cached_property
    def env(self):
        env = dict()
        paths = list()

        if 'PATH' in self.toolchain.env:
            paths.extend(self.toolchain.env['PATH'].split(os.pathsep))

        from dan.pkgconfig.package import get_cached_bindirs
        paths.extend([str(d) for d in get_cached_bindirs()])
        
        paths.extend(self.shared_dependencies_path)

        if len(paths):
            env['PATH'] = os.pathsep.join(paths)
        return env

    async def __initialize__(self):
        await super().__initialize__()

        previous_args = self.cache.get('link_args')
        if previous_args:
            args = self.toolchain.make_link_commands([obj.routput for obj in self.objs], self.output,
                                                     self._make_link_options())[0]
            args = [str(a) for a in args]
            if sorted(previous_args) != sorted(args):
                self.__dirty = True

    @cached_property
    def up_to_date(self):
        if self.__dirty:
            return False
        return super().up_to_date

    async def __build__(self):
        await super().__build__()

        # link
        self.info('linking %s...', self.output.name)
        try:
            commands, diags = await self.toolchain.link([obj.routput for obj in self.objs], self.output,
                                                        self._make_link_options())
            self.diagnostics.insert(diags, str(self.output))
        except LinkageFailure as err:
            self.diagnostics.insert(err.diags, str(self.output))
            err.target = self
            raise
        self.cache['link_args'] = [str(a) for a in commands[0]]
        self.debug('done')

    async def __install__(self, installer: Installer):
        await installer.install_bin(self.output)
        await super().__install__(installer)

    async def execute(self, *args, build=True, **kwargs):
        if build:
            await self.build()
        return await async_run([self.output, *args], logger=self, env=self.env, **kwargs)



import os
import sys
import contextlib

from dan import logging

from dan.core.find import find_file
from dan.core.pathlib import Path
from dan.core.terminal import TerminalMode, set_mode as set_terminal_mode, manager as term_manager

from dan.cli import click

from dan.core import diagnostics, asyncio
from dan.core.cache import Cache
from dan.core.settings import Settings
from dan.cxx.targets import Executable


from dan.make import InstallMode, Make
from dan.cli.vscode import Code


_minimal_options = [
    click.option('--build-path', '-B', help='Path where dan has been initialized.',
                 type=click.Path(resolve_path=True, path_type=Path), required=True, default='build', envvar='DAN_BUILD_PATH'),
]

_common_opts = [
    *_minimal_options,
    click.option('--quiet', '-q', is_flag=True,
                 help='Dont print informations (errors only).', envvar='DAN_QUIET'),
    click.option('--verbose', '-v', count=True,
                 help='Verbosity level.', envvar='DAN_VERBOSE'),
    click.option('--jobs', '-j',
                 help='Maximum jobs.', default=None, type=int, envvar='DAN_JOBS'),
    click.option('--no-status', is_flag=True,
                 help='Disable status', envvar='DAN_NOSTATUS'),
]


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func
    return _add_options


common_opts = add_options(_common_opts)
minimal_options = add_options(_minimal_options)


class CommandsContext:
    def __init__(self, *args, **kwds) -> None:
        self._make_args = [*args]
        self._make_kwds = {**kwds}
        self._make = None

    def update(self, *args, **kwds):
        if len(args):
            self._make_args.extend(*args)
        self._make_kwds.update(**kwds)
    
    @contextlib.asynccontextmanager
    async def __call__(self, *args, **kwargs):
        no_init = kwargs.pop('no_init', False)
        if kwargs.pop('no_status', False):
            kwargs['terminal_mode'] = TerminalMode.BASIC
        elif kwargs.pop('code', False):
            kwargs['terminal_mode'] = TerminalMode.CODE
        self.update(*args, **kwargs)
        quiet = self._make_kwds.pop('quiet', None)
        if quiet:
            self._make_kwds['verbose'] = -1
        if self._make is None:
            self._make = Make(*self._make_args, **self._make_kwds)
            if not no_init:
                await self._make.initialize()
        with self._make.context:
            yield self._make

    async def __aexit__(self, *exc):
        pass

pass_context = click.make_pass_decorator(CommandsContext)

@pass_context
def show_diags(ctx: CommandsContext):
    if diagnostics.enabled:
        diags = ctx._make.diagnostics
        if diags:
            click.echo(f'DIAGNOSTICS: {diags.to_json()}')

@click.group(no_args_is_help=True)
@click.version_option(package_name='dan-build')
@click.option('--quiet', '-q', is_flag=True,
              help='Dont print informations (errors only)')
@click.option('--verbose', '-v', count=True,
              help='Verbosity level')
@click.option('--jobs', '-j',
              help='Maximum jobs', default=None, type=int)
@click.pass_context
def cli(ctx: click.AsyncContext, **kwds):
    ctx.obj = CommandsContext(**kwds)
    ctx.call_on_close(show_diags)

@cli.command('cli', context_settings=dict(
    ignore_unknown_options=True,
    allow_extra_args=True,
))
@common_opts
@click.option('--help', is_flag=True)
@pass_context
@click.pass_context
async def user_cli_command(click_ctx, ctx, help, *args, **kwargs):
    async with ctx(no_init=False, **kwargs) as make:
        if not click_ctx.args:
            click.echo(user_cli.get_help(click_ctx))
            return 1
        args = click_ctx.args
        if help:
            args.append('--help')
        name, command, args = user_cli.resolve_command(click_ctx, click_ctx.args)
        setattr(click_ctx, 'obj', make)
        cmd_ctx = command.make_context(name, args, parent=click_ctx)
        return asyncio.may_await(cmd_ctx.invoke(command))
user_cli_command.add_help_option = False

@click.group()
def user_cli():
    pass
user_cli.context_class = click.AsyncContext

@cli.command()
@click.option('--verbose', '-v', count=True,
              help='Verbosity level')
@click.option('--toolchain', '-t', help='The toolchain to use',
              type=click.ToolchainParamType(), envvar='DAN_TOOLCHAIN')
@click.option('--setting', '-s', 'settings', help='Set or change a setting', multiple=True, type=click.SettingsParamType(Settings))
@click.option('--option', '-o', 'options', help='Set or change an option', multiple=True, type=click.OptionsParamType())
@click.option('--build-path', '-B', help='Path where dan has been initialized.',
              type=click.Path(resolve_path=True, path_type=Path), required=True, default='build', envvar='DAN_BUILD_PATH')
@click.option('--source-path', '-S', help='Path where source is located.',
              type=click.Path(resolve_path=True, path_type=Path), required=True, default='.', envvar='DAN_SOURCE_PATH')
@pass_context
async def configure(ctx: CommandsContext, toolchain: str, settings: tuple[str], options: tuple[str], source_path: Path, **kwds):
    """Configure dan project"""
    async with ctx(no_init=True, no_status=True, **kwds) as make:
        if toolchain is None and make.config.toolchain is None:
            from dan.cxx.detect import get_toolchains
            tp = click.Choice([*get_toolchains(create=False)["toolchains"].keys(), 'default'])
            toolchain = click.prompt('Toolchain', type=tp, default='default')

        await make.configure(source_path, toolchain)

        if len(settings):
            await make.apply_settings(*settings)

        # NOTE: intializing make after applying setting
        #       to check settings are valid implicitly (cache save skipped)
        await make.initialize()

        if len(options):
            await make.apply_options(*options)


@cli.command()
@click.option('--for-install', is_flag=True, help='Build for install purpose (will update rpaths [posix only])')
@common_opts
@click.option('--force', '-f', is_flag=True,
              help='Clean before building')
@click.argument('TARGETS', nargs=-1, type=click.TargetParamType())
@pass_context
async def build(ctx: CommandsContext, force=False, **kwds):
    """Build targets"""
    async with ctx(**kwds) as make:
        if force:
            await make.clean()
        await make.build()

@cli.command()
@common_opts
@click.option('--force', '-f', is_flag=True,
              help='Force re-install')
@click.argument('TARGETS', nargs=-1, type=click.TargetParamType())
@pass_context
async def install_dependencies(ctx: CommandsContext, force, **kwds):
    """Build targets"""
    async with ctx(**kwds) as make:
        await make.install_dependencies(force=force)

@cli.command()
@common_opts
@click.argument('MODE', type=click.Choice([v.name for v in InstallMode]), default=InstallMode.user.name)
@click.argument('TARGETS', nargs=-1, type=click.TargetParamType())
@pass_context
async def install(ctx: CommandsContext, mode: str, **kwargs):
    """Install targets"""
    async with ctx(**kwargs) as make:
        mode = InstallMode[mode]
        await make.install(mode)


@cli.command()
@common_opts
@click.option('--type', '-t', 'pkg_type', type=click.Choice(['tar.gz', 'zip']), default='tar.gz')
@click.argument('MODE', type=click.Choice([v.name for v in InstallMode]), default=InstallMode.user.name)
@click.argument('TARGETS', nargs=-1, type=click.TargetParamType())
@pass_context
async def package(ctx: CommandsContext, pkg_type, mode: str, **kwargs):
    """Package given targets"""
    async with ctx(**kwargs) as make:
        mode = InstallMode[mode]
        await make.package(pkg_type, mode)


@cli.command()
@click.option('--verbose', '-v', count=True,
              help='Verbosity level')
@click.option('--yes', '-y', is_flag=True, help='Proceed without asking')
@click.option('--root', '-r', help='Root path to search for installation manifest', type=click.Path(exists=True, file_okay=False))
@click.argument('NAME')
def uninstall(verbose: int, yes: bool, root: str, name: str):
    """Uninstall previous installation"""
    if verbose == 0:
        logging.getLogger().setLevel(logging.INFO)
    elif verbose == 1:
        logging.getLogger().setLevel(logging.DEBUG)
    else:
        logging.getLogger().setLevel(logging.TRACE)
    if root:
        paths = [root]
    else:
        paths = [
            '~/.local/share/dan',
            '/usr/local/share/dan',
            '/usr/share/dan',
        ]
    manifest = find_file(f'{name}-manifest.txt$', paths=paths)
    with open(manifest, 'r') as f:
        files = [(manifest.parent / mf.strip()).resolve()
                 for mf in f.readlines()]
    to_be_removed = '\n'.join([f" - {f}" for f in files])
    yes = yes or click.confirm(
        f'Following files will be removed:\n {to_be_removed}\nProceed ?')
    if yes:
        def rm_empty(dir: Path):
            if dir.is_empty:
                click.logger.debug(f'removing empty directory: {dir}')
                os.rmdir(dir)
                rm_empty(dir.parent)

        for f in files:
            click.logger.debug(f'removing: {f}')
            os.remove(f)
            rm_empty(f.parent)

        os.remove(manifest)
        rm_empty(manifest.parent)

@cli.group()
@pass_context
def ls(ctx: CommandsContext):
    """Inspect stuff"""
    ctx._make_kwds['terminal_mode'] = TerminalMode.BASIC

@ls.command()
@click.option('-a', '--all', 'all', is_flag=True, help='Show all targets (not only defaulted ones)')
@click.option('-t', '--type', 'show_type', is_flag=True, help='Show target\'s type')
@common_opts
@click.argument('TARGETS', nargs=-1)
@pass_context
async def targets(ctx: CommandsContext, all: bool, show_type: bool, **kwargs):
    """List targets"""
    kwargs['quiet'] = True
    async with ctx(**kwargs) as make:
        out = []
        for target in make.targets:
            if show_type:
                out.append(target.fullname + ' - ' + type(target).__name__)
            else:
                out.append(target.fullname)
        click.echo('\n'.join(out))


@ls.command()
@common_opts
@click.argument('TARGETS', nargs=-1)
@pass_context
async def tests(ctx: CommandsContext, **kwargs):
    """List tests"""
    kwargs['quiet'] = True
    async with ctx(**kwargs) as make:
        for t in make.tests:
            if len(t) > 1:
                for c in t.cases:
                    click.echo(f'{t.fullname}:{c.name}')
            else:
                click.echo(t.fullname)

@ls.command()
@common_opts
@click.argument('TARGETS', nargs=-1)
@pass_context
async def options(ctx: CommandsContext, **kwargs):
    """List tests"""
    kwargs['quiet'] = True
    async with ctx(**kwargs) as make:
        for o in make.all_options:
            current = ''
            if o.value != o.default:
                current = f', current: {o.value}'
            click.echo(f'{o.fullname}: {o.help} (type: {o.type.__name__}, default: {o.default}{current})')

@ls.command()
def toolchains(**kwargs):
    """List toolchains"""
    kwargs['quiet'] = True
    for name, _ in Make.toolchains()['toolchains'].items():
        click.echo(name)


@ls.command()
@common_opts
@click.option('-n', '--not-found', help='Show not-found dependencies', is_flag=True)
@click.argument('TARGET', type=click.TargetParamType(target_types=[Executable]))
@pass_context
async def runtime_dependencies(ctx: CommandsContext, not_found, target, **kwargs):
    """Inspect stuff"""
    async with ctx(**kwargs) as make:
        for t in make.root.all_targets:
            if t.fullname == target:
                break
        from dan.cxx import ldd
        for lib, lib_path in await ldd.get_runtime_dependencies(t):
            if lib_path or not_found:
                print(' ' * 7, lib, '=>', lib_path or 'not found')

@cli.command()
@common_opts
@click.argument('TARGETS', nargs=-1, type=click.TargetParamType())
@pass_context
async def clean(ctx, **kwargs):
    """Clean generated stuff"""
    async with ctx(**kwargs) as make:
        await make.clean()


@cli.command()
@common_opts
@click.argument('TARGETS', nargs=-1, type=click.TargetParamType())
@pass_context
async def run(ctx, **kwargs):
    """Run executable(s)"""
    async with ctx(**kwargs) as make:
        rc = await make.run()
        sys.exit(rc)


@cli.command()
@common_opts
@click.argument('TARGETS', nargs=-1, type=click.TargetParamType())
@pass_context
async def test(ctx, **kwargs):
    """Run tests"""
    async with ctx(**kwargs) as make:
        rc = await make.test()
        sys.exit(rc)


@cli.command()
@click.option('-s', '--script',
              help='Use a source script to resolve compilation environment')
@click.option('-p', '--path', 'paths',
              help='Use given path for compilers lookup', multiple=True, type=click.Path(exists=True, file_okay=False))
@click.option('--verbose', '-v', count=True,
              help='Pring debug informations.', envvar='DAN_VERBOSE')
def scan_toolchains(script: str, paths: list[str], verbose, **kwargs):
    """Scan system toolchains"""
    match verbose:
        case 1:
            log_level = logging.DEBUG
        case 2:
            log_level = logging.TRACE
        case -1:
            log_level = logging.ERROR
        case 0:
            log_level = logging.INFO
        case _:
            logging.getLogger().warning('unknown verbosity level: %s, using INFO', verbose)
            log_level = logging.INFO
    logging.getLogger().setLevel(log_level)
    from dan.cxx.detect import create_toolchains, load_env_toolchain
    if script:
        load_env_toolchain(script)
    else:
        create_toolchains(paths if len(paths) else None)

@cli.command()
@common_opts
@click.argument('TARGETS', nargs=-1, type=click.TargetParamType())
@pass_context
async def env(ctx: CommandsContext, **kwds):
    """Show environment."""
    kwds['quiet'] = True
    async with ctx(**kwds) as make:
        for k, v in make.env.items():
            click.echo(f'{k}={v}')


@cli.command()
@common_opts
@click.argument('TARGETS', nargs=-1, type=click.TargetParamType())
@pass_context
async def shell(ctx: CommandsContext, **kwds):
    """Open a new shell with suitable environment."""
    from dan.core.runners import sync_run
    from copy import copy
    # kwds['quiet'] = True
    async with ctx(**kwds) as make:
        env = dict(os.environ)
        for k, v in make.env.items():
            env[k] = v

        click.logger.info('entering dan shell...')
        click.logger.debug('env: %s', env)
        
        sync_run('bash', cwd=make.root.build_path, env=env, pipe=False)


@cli.group()
def code():
    """VS-Code specific commands"""


# from dan.core.bench import benchmark, report_all

@code.command()
@common_opts
@click.argument('TARGETS', nargs=-1)
@pass_context
async def get_targets(ctx: CommandsContext, **kwargs):
    kwargs.update({'quiet': True, 'diags': True, 'no_status': True})
    # with benchmark('get-targets') as bench:
        # bench.begin('make')
    async with ctx(**kwargs) as make:
            # bench.end()
        out = []
        targets = make.context.root.all_targets
            # with bench('load-dependencies'):
        async with asyncio.TaskGroup() as g:
            for target in targets:
                g.create_task(target.load_dependencies())
            # with bench('gen-output'):
        for target in targets:
            # with bench(f'gen-output-{target.name}'):
                out.append({
                    'name': target.name,
                    'fullname': target.fullname,
                    'buildPath': str(target.build_path),
                    'srcPath': str(target.source_path),
                    'output': str(target.output),
                    'executable': isinstance(target, Executable),
                    'type': type(target).__name__,
                    'env': target.env if isinstance(target, Executable) else None,
                })
            # with bench('json-dump'):
        import json
        click.echo(json.dumps(out))
    # report_all()

@code.command()
@common_opts
@click.argument('TARGETS', nargs=-1)
@pass_context
async def get_tests(ctx: CommandsContext, **kwargs):
    kwargs.update({'quiet': True, 'diags': True, 'no_status': True})
    async with ctx(**kwargs) as make:
        import json
        out = list()
        for t in make.context.root.all_tests:
            out.append(t.fullname)
            if len(t) > 1:
                for c in t.cases:
                    out.append(f'{t.fullname}:{c.name}')
        click.echo(json.dumps(out))


@code.command()
@common_opts
@click.option('--pretty', is_flag=True)
@click.argument('TARGETS', nargs=-1)
@pass_context
async def get_test_suites(ctx: CommandsContext, pretty, **kwargs):
    kwargs.update({'quiet': True, 'diags': True, 'no_status': True})
    async with ctx(**kwargs) as make:
        code = Code(make)
        click.echo(code.get_test_suites(pretty))


@code.command()
def get_toolchains(**kwargs):
    import json
    click.echo(json.dumps(list(Make.toolchains()['toolchains'].keys())))


@code.command()
@click.option('--for-install', is_flag=True, help='Build for install purpose (will update rpaths [posix only])')
@common_opts
@click.option('--force', '-f', is_flag=True,
              help='Clean before building')
@click.argument('TARGETS', nargs=-1, type=click.TargetParamType())
@pass_context
async def build(ctx: CommandsContext, force=False, **kwargs):
    """Build targets (vscode version)"""
    async with ctx(**kwargs, diags=True, code=True) as make:
        if force:
            await make.clean()
        await make.build()


@code.command()
@minimal_options
@click.argument('SOURCES', nargs=-1, type=click.Path(exists=True, dir_okay=False, resolve_path=True))
@pass_context
async def get_source_configuration(ctx: CommandsContext, sources, **kwargs):
    kwargs.update({'quiet': True, 'diags': True, 'no_status': True})
    async with ctx(**kwargs) as make:
        code = Code(make)
        click.echo(await code.get_sources_configuration(sources))


@code.command()
@minimal_options
@pass_context
async def get_workspace_browse_configuration(ctx: CommandsContext, **kwargs):
    kwargs.update({'quiet': True, 'diags': True, 'no_status': True})
    async with ctx(**kwargs) as make:
        code = Code(make)
        click.echo(await code.get_workspace_browse_configuration())

@cli.result_callback()
@pass_context
async def process_result(ctx, result, **kwargs):
    await Cache.save_all()

def main():
    import sys
    try:
        loop = asyncio.new_event_loop()
        cli(auto_envvar_prefix='DAN')
    except Exception as err:
        click.logger.error(str(err))
        _ex_type, _ex, tb = sys.exc_info()
        import traceback
        click.logger.debug(' '.join(traceback.format_tb(tb)))
        try:
            # wait asyncio loop to terminate
            loop.run_until_complete()
        except Exception:
            pass
        asyncio.run(Cache.save_all())
        return -1
    finally:
        term = term_manager()
        term.stop()
        if term._thread:
            term._thread.get_loop().run_until_complete(term._thread)




import os
from dan.core.pathlib import Path
import re
import typing as t

include_paths_lookup = [
    '~/.local/include',
    '/usr/local/include',
    '/usr/include',
]

programs_paths_lookup = os.getenv('PATH', '').split(os.pathsep)

library_paths_lookup = [
    '$LD_LIBRARY_PATH'
    '~/.local/lib',
    '~/.local/lib64',
    '/usr/local/lib',
    '/usr/local/lib64',
    '/usr/lib',
    '/usr/lib/lib64',
    '/lib',
    '/lib64',
]


def find_file(expr, paths, flags = 0) -> Path:
    r = re.compile(expr, flags)
    if isinstance(paths, (str, Path)):
        paths = [paths]
    for path in paths:
        for root, _, files in os.walk(os.path.expandvars(os.path.expanduser(path))):
            for file in files:
                if r.match(file):
                    return Path(root) / file


def find_files(expr, paths, flags = 0) -> t.Generator[Path, None, None]:
    r = re.compile(expr, flags)
    if isinstance(paths, (str, Path)):
        paths = [paths]
    for path in paths:
        for root, _, _files in os.walk(os.path.expandvars(os.path.expanduser(path))):
            for file in _files:
                if r.match(file):
                    yield (Path(root) / file)


def find_include_path(name, paths: list[str|Path] = None) -> Path:
    paths = paths or list()
    return find_file(name, [*paths, *include_paths_lookup])


def find_library(name, paths: list[str|Path] = None) -> Path:
    paths = paths or list()
    if os.name == 'posix':
        expr = fr'lib{name}\.(so|a)'
    elif os.name == 'nt':
        expr = fr'lib{name}\.(lib|dll)'
    return find_file(expr, [*paths, *library_paths_lookup])

def find_executable(name, paths: list[str|Path] = None, default_paths=True) -> Path:
    paths = paths or list()
    if os.name == 'posix':
        expr = name + '$'
    elif os.name == 'nt':
        expr = f'{name}.exe$'
    if default_paths:
        paths.extend(programs_paths_lookup)
    return find_file(expr, paths)

def find_executables(name, paths: list[str|Path] = None, default_paths=True) -> t.Generator[Path, None, None]:
    paths = paths or list()
    if os.name == 'posix':
        expr = name + '$'
    elif os.name == 'nt':
        expr = f'{name}.exe$'
    if default_paths:
        paths.extend(programs_paths_lookup)
    yield from find_files(expr, paths)
